////////////////
//This file contains only class Plate (since it needs to be made Public)

package FACS.Annotator;

import java.io.Serializable;
import java.util.Vector;
import java.util.LinkedList;
import java.util.Hashtable;
import java.awt.Dimension;
import javax.swing.event.*;
import javax.swing.*;

/**
 * Plate describes one (e.g. 96-well) plate of a FACS experiment
 * @author Mark
 *
 */
public class Plate implements Serializable {
	private static final long serialVersionUID = 1L;
	public Vector data;
	public int rows, cols;
	public String name;
	public int num;
	public boolean fillAcross;
	
	transient JList pl; //will not be serialized
	transient JTabbedPane tabPlates;

	protected transient javax.swing.event.EventListenerList listenerList = new javax.swing.event.EventListenerList();

	public Plate() {
		pl = null;
		tabPlates = null;
		data = null;
		rows = 0;
		cols = 0;
		name = null;
		num = 0;
		fillAcross = true;
	}

	public void setListenerList() {
		listenerList = new javax.swing.event.EventListenerList();
	}

	public Plate(JList pl, JTabbedPane tabPlates, int num) {
		this.pl = pl;
		this.tabPlates = tabPlates;
		data = new Vector();
		rows = 8;
		cols = 12;
		name = "Samples";
		this.num = num;
		fillAcross = true;
	}

	public Plate(JList pl, JTabbedPane tabPlates, String name, int num) {
		this.pl = pl;
		this.tabPlates = tabPlates;
		data = new Vector();
		rows = 8;
		cols = 12;
		this.name = name;
		this.num = num;
		fillAcross = true;
	}

	public int getRows() {
		return rows;
	}

	public void setRows(int r) {
		rows = r;
	}

	public int getCols() {
		return cols;
	}

	public void setCols(int c) {
		cols = c;
	}

	public boolean isFillAcross() {
		return fillAcross;
	}

	public void setFillAcross(boolean b) {
		fillAcross = b;
	}

	/**
	 * @return Vector of all files in the plate
	 */
	public Vector getData() {
		return data;
	}

	public void setData(Vector v) {
		data = v;
	}

	public void setName(String n) {
		name = n;
	}

	/**
	 * @return Number of the plate in its series,
	 * (e.g. 2 in "Samples 2")
	 */
	public int getNum() {
		return num;
	}

	public void setNum(int n) {
		num = n;
	}

	public void setSize(int rows, int cols) {
		this.rows = rows;
		this.cols = cols;
	}

	/**
	 * @return Plate base name plus its number
	 * (e.g. "Samples 2")
	 */
	public String getFullName() {
		return name+" "+num;
	}

	public String toString() {
		return getFullName();
	}

	/**
	 * @return Plate base name (e.g. "Samples")
	 */
	public String getName() { //for XML encoder
		return name;
	}

	/**
	 * Public accessor method to change plate size - calls up an input dialog,
	 * changes size, then fires a ChangeEvent so that the table will know that the size has changed.
	 */
	public void resizePlate() {
		PlateDialog pd = new PlateDialog(rows,cols,name);
		name = pd.name;
		if(pd.rows*pd.cols >= data.size()) {
			rows = pd.rows; cols = pd.cols;
		}
		int i = ((DefaultListModel)pl.getModel()).indexOf(this);
		((DefaultListModel)pl.getModel()).setElementAt(this,i); //fireContentsChanged is protected in DefaultListModel
		fireChangeEvent(new ChangeEvent(this)); //for plate table
	}

	public void add(LinkedList ll) {
		while(data.size()+1 <= rows*cols && ll.size() > 0) {
			data.addElement(ll.removeFirst());
		}
		fireChangeEvent(new ChangeEvent(this)); //for plate table
		if(ll.size() > 0) { //create a new Plate for excess
			Plate newP = new Plate(pl, tabPlates, name, num+1);
			newP.setSize(rows,cols);
			((DefaultListModel)pl.getModel()).addElement(newP);
			tabPlates.addTab(newP.getFullName(),new NullComponent());
			tabPlates.setPreferredSize(new Dimension(100,30));
			newP.add(ll);
		}
	}

	//changeevent handling procs
	public void addChangeListener(ChangeListener listener) {
		listenerList.add(ChangeListener.class, listener);
	}

	public void removeChangeListener(ChangeListener listener) {
		listenerList.remove(ChangeListener.class, listener);
	}

	// This private class is used to fire events
	void fireChangeEvent(ChangeEvent ce) {
		Object[] listeners = listenerList.getListenerList();
		// Each listener occupies two elements - the first is the listener class
		// and the second is the listener instance
		for (int i=0; i<listeners.length; i+=2) {
			if (listeners[i]==ChangeListener.class) {
				((ChangeListener)listeners[i+1]).stateChanged(ce);
			}
		}
	}

	public void transpose() {
		fillAcross = !fillAcross;
	}

	public void insert(Hashtable fd, int n) {
		data.insertElementAt(fd,n);
	}

	public void setBlank(int i) {
		if(i >= data.size()) //deleted a non-existent file
			return;
		data.setElementAt(null,i);
	}

	public void deleteBlanks() {
		for(int i = 0; i < data.size(); i++) {
			//if(data.elementAt(i) instanceof Blank) {
			//	if(((Blank)data.elementAt(i)).index == 0) {
				if(data.elementAt(i) == null) {
					data.removeElementAt(i);
					i--;
				}
			//}
		}
	}

	/**
	 * @param row
	 * @param col
	 * @return index in files Vector ('data') of a particular cell in the plate/table
	 */
	public int getIndex(int row, int col) { //get index in data vector of a particular cell
		int n;
		if(fillAcross)
			n = (row)*cols+col;
		else
			n = rows*col+row;
		return n;
	}

	/**
	 * @param index
	 * @return [row, column] of a particular index in the files Vector
	 */
	public int[] getRC(int index) { //inverse of getIndex
		int[] rc = new int[2]; //row, col
		if(fillAcross) {
			rc[1] = index % cols;
			rc[0] = (index-rc[1]) / cols;
		} else {
			rc[0] = index % rows;
			rc[1] = (index-rc[0]) / rows;
		}
		return rc;
	}

	public String getFileName(int row, int col) {
		return (String)data.elementAt(getIndex(row,col));
	}
}