/////////////////////
//This file contains PlateTableModel, table transferhandlers, PlateDialog, and RowHeader

package FACS.Annotator;

import javax.swing.*;
import java.awt.event.*;
import javax.swing.event.*;
import java.util.*;
import javax.swing.table.*;

public class PlateTableModel extends AbstractTableModel implements ChangeListener {
	private Plate plate;
	boolean fillAcross;
	int rows, cols;
	//String[] keys; //displayed keys for FACSData.annotations
	private String key;
	private String[] keys; //for use in getValuesAt

	public PlateTableModel(Plate plate) {
		this.plate = plate;
		fillAcross = plate.isFillAcross();
		rows = plate.getRows();
		cols = plate.getCols();
		/*keys = new String[1];
		keys[0] = "filename";*/
		key = "filename";
		keys = new String[1]; keys[0] = key;
		plate.addChangeListener(this);
	}

	public int getColumnCount() {
		return cols;
	}

	public int getRowCount() {
		return rows;
	}

	public String getColumnName(int col) {
		return (col+1)+"";
	}

	public String getRowName(int row) {
		return row+"";
	}

	public void setKey(String key) {
		this.key = key;
		fireTableDataChanged();
	}

	public void setKeys(String[] keys) {
		this.keys = keys;
	}

	public boolean isFillAcross() {
		return plate.isFillAcross();
	}

	public boolean isFileMode() {
		return key.equals("filename");
	}

	public int getIndex(int row, int col) { //get index in data vector of a particular cell
		int n;
		if(plate.isFillAcross())
			n = (row)*cols+col;
		else
			n = rows*col+row;
		return n;
	}

	public Object getValueAt(int row, int col) {
		return getValueAt(row, col, key);
	}

	public Object getValueAt(int row, int col, String key) {
		int n = getIndex(row,col);
		if(n < plate.getData().size()) {
			Hashtable fd = (Hashtable)plate.getData().elementAt(n);
			if(fd == null || fd.isEmpty())
				return null;

			/*String val = "";
			for(int i = 0; i < keys.length; i++) {
				val += (String)fd.getAnnotation(keys[i]);
				if(i != keys.length-1)
					val += ", ";
			}
			return val;*/
			return (String)fd.get(key);
		} else {
			return null;
		}
	}

	public String getValuesAt(int row, int col) { //for copy
		String o = "";
		for(int i = 0; i < keys.length; i++) {
			o += keys[i]+"|"+(String)getValueAt(row,col,keys[i]);
			if(i < keys.length-1)
				o += "|";
		}
		return o;
	}

	public Hashtable getFDAt(int row, int col) {
		int n = getIndex(row,col);
		if(n < plate.getData().size()) {
			return (Hashtable)plate.getData().elementAt(n);
		} else {
			return null;
		}
	}

	public Class getColumnClass(int c) {
		return String.class;
	}

	public boolean isCellEditable(int row, int col) {
		/*if(keys.length > 1 || keys[0].equals("filename"))
			return false;*/
		if(key.equals("filename"))
			return false;
		int n = getIndex(row,col);
		if(n >= plate.getData().size())
			return false;
		if(((Hashtable)plate.getData().elementAt(n)) == null)
			return false;
		return true;
	}

	public void setValueAt(Object value, int row, int col, String key) {
		int n = getIndex(row,col);
		if(n < plate.getData().size()) {
			Hashtable fd = (Hashtable)plate.getData().elementAt(n);
			fd.put(key,value);
			fireTableCellUpdated(row, col);
		}
	}

	public void setValueAt(Object value, int row, int col) {
		setValueAt(value, row, col, key);
	}

	public void setValuesAt(String values, int row, int col) {
		if(values.indexOf("|") != -1) {
			String[] pcs = values.split("\\|");
			for(int i = 0; i < pcs.length; i++) {
				if(!pcs[i].equals("filename")) //we'll not let them change filename at the same time as annotations
					setValueAt(pcs[i+1],row,col,pcs[i]);
				i++;
			}
		} else { //pasting from Excel or something... (use current key as default)
			setValueAt(values,row,col);
		}
	}

	public void stateChanged(ChangeEvent ce) {
		int oldcols = cols;
		rows = plate.getRows();
		cols = plate.getCols();
		if(oldcols != cols)
			fireTableChanged(new TableModelEvent(this, TableModelEvent.HEADER_ROW));
		fireTableDataChanged();
	}

	public void transpose() {
		plate.transpose();
		fillAcross = plate.isFillAcross();
		fireTableDataChanged();
	}

	public void insert(Hashtable fd, int row, int col) {
		if(plate.getData().size() == plate.getRows()*plate.getCols())
			return;
		int n = getIndex(row,col);
		plate.insert(fd, n);
		fireTableDataChanged();
	}

	public void insert(LinkedList ll, int row, int col) {
		int n = getIndex(row,col);
		while(ll.size() > 0) {
			plate.insert((Hashtable)ll.removeFirst(),n++); //"n++" evaluates before increment
		}
		fireTableDataChanged();
	}

	public Vector remove(int[] rows, int[] cols, int offset1, int offset2) {
		//if offset = 0, was not from a DnD -> replace in files list!
		Vector out = new Vector();
		//offset1 = insertion site, offset2 = insertion length
		offset1--; //insert _at_and_after_ offset1, so need to decrement
		int index;
		if(fillAcross) {
			for(int i = 0; i < rows.length; i++) {
				for(int j = 0; j < cols.length; j++) {
					if(offset1 < (index=getIndex(rows[i],cols[j])))
						index += offset2;
					if(offset1 == -1)
						out.add(getFDAt(rows[i],cols[j]));
					plate.setBlank(index);
				}
			}
		} else {
			for(int j = 0; j < cols.length; j++) {
				for(int i = 0; i < rows.length; i++) {
					if(offset1 < (index=getIndex(rows[i],cols[j])))
						index += offset2;
					if(offset1 == 0)
						out.add(getFDAt(rows[i],cols[j]));
					plate.setBlank(index);
				}
			}
		}
		plate.deleteBlanks();
		fireTableDataChanged();
		return out;
	}
}

class TableTransferHandler extends StringTransferHandler {
    private int[] rows = null;
	private int[] cols = null;
    private int addIndex = -1; //Location where items were added
    private int addCount = 0;  //Number of items added.

	Hashtable files;

	public TableTransferHandler(Hashtable files) {
		this.files = files;
	}

    protected String exportString(JComponent c) {
		//CAN PASTE INTO NON-CONSECUTIVE, BUT NOT COPY FROM NON-CONSECUTIVE (no account for skipped columns)
        JTable table = (JTable)c;
		PlateTableModel model = (PlateTableModel)table.getModel();
        rows = table.getSelectedRows();
		cols = table.getSelectedColumns();

		//fix rows or cols if row-only or col-only mode (thus, value given as 0)
		if(model.isFillAcross()) { //can select rows
			if(cols.length == 0) {
				cols = incrementArray(table.getColumnCount());
			}
		} else {
			if(rows.length == 0) {
				rows = incrementArray(table.getRowCount());
			}
		}

        StringBuffer buff = new StringBuffer();
		if(model.isFillAcross() || !model.isFileMode()) {
			for(int i = 0; i < rows.length; i++) {
				for(int j = 0; j < cols.length; j++) {
					if(model.isFileMode())
						buff.append(model.getValueAt(rows[i],cols[j]));
					else
						buff.append(model.getValuesAt(rows[i],cols[j])); //can have multiple key-value pairs
					if (j != cols.length - 1) {
						buff.append("\t");
					}
				}
				if (!model.isFileMode() && (i != rows.length - 1)) { //filemode will always go fill-across, for convenience
					buff.append("\n");
				}
			}
		} else {
			for(int j = 0; j < cols.length; j++) {
				for(int i = 0; i < rows.length; i++) {
					buff.append(table.getValueAt(rows[i],cols[j]));
					if (i != rows.length - 1) {
						buff.append("\t");
					}
				}
			}
		}
        return buff.toString();
    }

	public static int[] incrementArray(int max) {
		int[] out = new int[max];
		for(int i = 0; i < out.length; i++) {
			out[i] = i;
		}
		return out;
	}

	protected boolean importString(JComponent c, String str) {
		JTable target = (JTable)c;
        PlateTableModel model = (PlateTableModel)target.getModel();
		if(model.isFileMode())
			return importStringF(c,str);
		else
			return importStringP(c,str);
	}

    protected boolean importStringF(JComponent c, String str) {
        JTable target = (JTable)c;
        PlateTableModel model = (PlateTableModel)target.getModel();

		//is imported selection allowed?
		if(rows.length == 0 && cols.length == 0) //only allowing drag-n-drop from this table
			return false;
		if(model.isFillAcross()) {
			//can only select a single incomplete row of data or multiple full rows of data
			//[cols length given as zero if row-only mode...]
			if(cols.length < target.getColumnCount() && !(cols.length == 0)) {
				if(rows.length > 1) {
					return false;
				}
			}
		} else {
			//can only select a single incomplete column or multiple complete columns
			if(rows.length < target.getRowCount() && !(rows.length == 0)) {
				if(cols.length > 1) {
					return false;
				}
			}
		}
	
        int row = target.getSelectedRow();
		int col = target.getSelectedColumn();

        /*//Prevent the user from dropping data back on itself.
        //For example, if the user is moving rows #4,#5,#6 and #7 and
        //attempts to insert the rows after row #5, this would
        //be problematic when removing the original rows.
        //So this is not allowed.
        if (rows != null && index >= rows[0] - 1 &&
              index <= rows[rows.length - 1]) {
            rows = null;
            return;
        }

        int max = model.getRowCount();
        if (index < 0) {
            index = max;
        } else {
            index++;
            if (index > max) {
                index = max;
            }
        }*/
        addIndex = model.getIndex(row,col);
        String[] values = str.split("\t");
        addCount = values.length;
		LinkedList ll = TableTransferHandler.getFDsFromName(files,values);
		model.insert(ll, row, col);
		return true;
    }
    
    public static LinkedList getFDsFromName(Hashtable annos, Object[] name) {
		LinkedList ll = new LinkedList();
		for(int n = 0; n < name.length; n++) {
			if(((String)name[n]).equals("") || !annos.containsKey((String)name[n])) {
				//ll.addLast(new Blank(1));
				ll.addLast(new Hashtable());
			} else {
				ll.addLast(annos.get((String)name[n]));
			}
		}
		return ll;
	}

	protected boolean importStringP(JComponent c, String str) { //paste annotations
		//CAN PASTE INTO NON-CONSECUTIVE, BUT NOT COPY FROM NON-CONSECUTIVE (no account for skipped columns)
		JTable target = (JTable)c;
        PlateTableModel model = (PlateTableModel)target.getModel();

		int[] rows2 = target.getSelectedRows();
		int[] cols2 = target.getSelectedColumns();

		String[] values1 = str.split("\n");
		String[][] matrix = new String[values1.length][];
		String[] values2;
		for(int i = 0; i < values1.length; i++) {
			values2 = values1[i].split("\t");
			matrix[i] = values2;
		}
		
		if(matrix.length == 0 || matrix[0].length == 0)
			return false;
		if(rows2.length > matrix.length || cols2.length > matrix[0].length) { //selection is larger than imported data
			//will use mod to fill out completely
			int maxR, maxC;
			maxR = matrix.length;
			for(int i = 0; i < rows2.length; i++) {
				maxC = matrix[i % maxR].length;
				for(int j = 0; j < cols2.length; j++) {
					model.setValuesAt(matrix[i % maxR][j % maxC],rows2[i],cols2[j]);
				}
			}
		} else { //selection smaller than/equal to imported data
			boolean singlecell = (rows2.length == 1) && (cols2.length == 1);
			boolean equal = (rows2.length == matrix.length) && (cols2.length == matrix[0].length);
			if(singlecell || equal) { //only allow to paste into a single cell (will auto. fill out, using that as upper left)
					//or into an equal sized selection
				for(int i = 0; i < matrix.length; i++) {  //will be rows
					for(int j = 0; j < matrix[i].length; j++) { //will be cols
						model.setValuesAt(matrix[i][j],rows2[0]+i,cols2[0]+j);
					}
				}
			} else { 
				return false;
			}
		}
		return true;
	}

    protected void cleanup(JComponent c, boolean remove) {
        JTable source = (JTable)c;
        if (remove && addCount > 0) { //need to get rid of entries
            PlateTableModel model = (PlateTableModel)source.getModel();
			if(model.isFileMode())
	            model.remove(rows,cols,addIndex,addCount);
        }
        rows = new int[0]; cols = new int[0];
        addCount = 0;
        addIndex = -1;
    }
}

class PlateDialog extends JDialog implements ActionListener {
	public int rows, cols;
	public String name;

	JTextField tfRows, tfCols, tfName;

	public PlateDialog(int rows, int cols, String name) {
		super((JFrame)null,"Resize Plate...",true);
		this.rows = rows;
		this.cols = cols;

		getContentPane().setLayout(new BoxLayout(getContentPane(), BoxLayout.Y_AXIS));

		JPanel p0 = new JPanel();
		p0.add(new JLabel("Name:"));
		tfName = new JTextField(name,20);
		p0.add(tfName);
		getContentPane().add(p0);
	
		JPanel p1 = new JPanel();
		p1.add(new JLabel("Rows:"));
		tfRows = new JTextField(rows+"",10);
		p1.add(tfRows);
		p1.add(new JLabel("Cols:"));
		tfCols = new JTextField(cols+"",10);
		p1.add(tfCols);
		getContentPane().add(p1);

		JButton bOk = new JButton("OK");
		bOk.addActionListener(this);
		bOk.setActionCommand("okay");
		getContentPane().add(bOk);

		validate(); pack();

		setVisible(true);
	}

	public void actionPerformed(ActionEvent e) {
		if(e.getActionCommand().equals("okay")) {
			try {
				int a = new Integer(tfRows.getText()).intValue();
				if(a < 0)
					throw new NumberFormatException();
				rows = a;
			} catch (NumberFormatException nfe) {}
			try {
				int a = new Integer(tfCols.getText()).intValue();
				if(a < 0)
					throw new NumberFormatException();
				cols = a;
			} catch (NumberFormatException nfe) {}
			name = tfName.getText();
			setVisible(false);
		}
	}
}

class RowHeaderRenderer extends JLabel implements ListCellRenderer {
   
	public RowHeaderRenderer(JTable table) {
		JTableHeader header = table.getTableHeader();
		setOpaque(true);
		setBorder(UIManager.getBorder("TableHeader.cellBorder"));
		setHorizontalAlignment(CENTER);
		setForeground(header.getForeground());
		setBackground(header.getBackground());
		setFont(header.getFont());
	}
  
	public java.awt.Component getListCellRendererComponent(JList list, Object value, int index, boolean isSelected, boolean cellHasFocus) {
		setText((value == null) ? "" : value.toString());
		return this;
	}
}
