//Class that describes a node in a hierarchical cluster

package FACS;

import java.util.Vector;

public class Cluster implements java.io.Serializable {
	Cluster child1, child2; //for int. nodes only
	String name; double[] vals; //for end nodes only
	int depth;

	public Cluster(String name, double[] vals) { //end node
		child1 = null;
		child2 = null;
		this.name = name;
		this.vals = vals;
		depth = 0;
	}

	public Cluster(Cluster child1, Cluster child2) { //top or intermediate node
		this.child1 = child1;
		child1.incrementDepth();
		this.child2 = child2;
		child2.incrementDepth();
		name = null;
		vals = null;
	}

	public Cluster getChild1() {
		return child1;
	}

	public Cluster getChild2() {
		return child2;
	}

	public int getDepth() {
		return depth;
	}

	public String getName() {
		return name;
	}

	public double[] getVals() {
		return vals;
	}

	public void setChild1(Cluster c) {
		child1 = c;
	}
	
	public void setChild2(Cluster c) {
		child2 = c;
	}

	public void setName(String name) {
		this.name = name;
	}

	public void setVals(double[] v) {
		vals = v;
	}

	public void setDepth(int d) {
		depth = d;
	}

	public boolean isEnd() {
		return child1 == null;
	}

	public void incrementDepth() {
		if(isEnd())
			depth++;
		else {
			child1.incrementDepth();
			child2.incrementDepth();
		}
	}

	public Vector getAllChildren(Vector children) { //[recusively] return all end nodes (for use in d())
		if(isEnd()) {
			children.add(this);
		} else {
			child1.getAllChildren(children);
			child2.getAllChildren(children);
		}
		return children;
	}

	public String toString() {
		if(isEnd())
			return "<table><tr><td width=100%>"+name+"</td>"+printVals()+"</tr></table>";
		return "<ul><li>"+child1.toString()+"</li><li>"+child2.toString()+"</li></ul>";
	}

	public String printVals() {
		if(!isEnd())
			return "";
		StringBuffer b = new StringBuffer();
		b.append("");
		for(int i = 0; i < vals.length; i++) {
			b.append("<td>"+round(vals[i],2)+"</td>");
		}
		b.append("");
		return b.toString();
	}

	public String round(double in, int dec) {
		in *= Math.pow(10,dec);
		double out = ((double)((int)in))/Math.pow(10,dec);
		return new Double(out).toString();
	}

	public static double dist(Cluster n1, Cluster n2) { //distance between two end node clusters
		//implement as the angle between the two vals vetors
		if(!n1.isEnd() || !n2.isEnd())
			return Double.NaN;

		return dotProd(n1.getVals(),n2.getVals()) / ( norm(n1.getVals())*norm(n2.getVals()) );
	}

	public static double dotProd(double[] one, double[] two) {
		if(one.length != two.length)
			return Double.NaN;
		double sum = 0;
		for(int i = 0; i < one.length; i++) {
			sum += one[i]*two[i];
		}
		return sum;
	}

	public static double norm(double[] v) {
		return Math.sqrt(dotProd(v,v));
	}

	public static double d(Cluster c1, Cluster c2) { //distance between clusters - implemented as average distance between end nodes
		Vector children1 = new Vector();
		c1.getAllChildren(children1);
		Vector children2 = new Vector();
		c2.getAllChildren(children2);

		Cluster n1, n2;
		double sum = 0;
		for(int i = 0; i < children1.size(); i++) {
			for(int j = 0; j < children2.size(); j++) {
				n1 = (Cluster)children1.elementAt(i);
				n2 = (Cluster)children2.elementAt(j);
				sum += dist(n1,n2);
			}
		}
		sum /= (double)(children1.size()*children2.size());
		return sum;
	}
}
