//FACSData class... used to store data contained in FCS files (with or without actual data points)

package FACS;

import java.util.Hashtable;
import java.util.Set;
import java.io.File;
import java.io.Serializable;

public class FACSData implements Serializable {

	private static final long serialVersionUID = 1L;
	public Hashtable text;
    public int[][] data;
    public FParameter[] params;
    public String name;
    public String filename;
    public File f;
    public int index;
	public Hashtable annotations;
	public boolean comp;
    
    public FACSData(int index, String filename, String name, Hashtable text, FParameter[] params) {
        this.index = index;
        this.filename = filename;
        this.name = name;
        
        this.text = text;
        this.params = params;
        
        this.name = getName();

		annotations = getDefaultAnnotations();

		comp = false;
		data = new int[0][0];
    }

	public FACSData(int index, File f, String name, Hashtable text, FParameter[] params, int[][] data, boolean comp) {
        this.index = index;
        this.f = f;
        this.name = name;
        filename = f.getName();
        this.text = text;
        this.data = data;
        this.params = params;
		this.comp = comp;
        
        this.name = getName();

		annotations = getDefaultAnnotations();
    }

	public void setAnnotations(Hashtable annotations) {
		if(annotations == null) //don't overwrite if we don't actually get annotations
			return;

		this.annotations = annotations;

		for(int i = 0; i < params.length; i++) {
			if(annotations.containsKey("P"+(i+1)))
				params[i].lname = (String)annotations.get("P"+(i+1));
		}

		this.name = getName(annotations);
	}

	public FACSData(int index) { //just for class Blank
		this.index = index;
	}

	public String getName() {
		if(name != null)
			return name;
		if(annotations != null)
			return getName(annotations);
		return getName(text);
	}
    
    public String getName(Hashtable annos) {
		if(annos.containsKey("SAMPLE ID"))
			return (String)annos.get("SAMPLE ID");
		if(annos.containsKey("CELLS"))
			return (String)annos.get("CELLS");
		if(annos.containsKey("TUBE NAME"))
			return (String)annos.get("TUBE NAME");
		if(annos.containsKey("COM"))
			return (String)annos.get("COM");
		if(annos.containsKey("SMNO"))
			return (String)annos.get("SMNO");
		if(annos.containsKey("SRC"))
			return (String)annos.get("SRC");
		if(annos.containsKey("EXPERIMENT NAME"))
			return (String)annos.get("EXPERIMENT NAME");
		if(annos.containsKey("PROJ"))
			return (String)annos.get("PROJ");
        return filename;
    }

	public Hashtable getDefaultAnnotations() {
		Hashtable h = new Hashtable();
		h.put("filename",filename);
		if(text.containsKey("SAMPLE ID"))
            h.put("SAMPLE ID",text.get("SAMPLE ID"));
        if(text.containsKey("CELLS"))
            h.put("CELLS",text.get("CELLS"));
        if(text.containsKey("COM"))
            h.put("COM",text.get("COM"));
        if(text.containsKey("SMNO"))
            h.put("SMNO",text.get("SMNO"));
        if(text.containsKey("SRC"))
            h.put("SRC",text.get("SRC"));
        if(text.containsKey("PROJ"))
            h.put("PROJ",text.get("PROJ"));
        if(text.containsKey("TUBE NAME"))
            h.put("TUBE NAME",text.get("TUBE NAME"));
        if(text.containsKey("EXPERIMENT NAME"))
            h.put("EXPERIMENT NAME",text.get("EXPERIMENT NAME"));
		for(int i = 0; i < params.length; i++) {
			h.put("P"+(i+1), params[i].lname);
		}
		return h;
	}

	/*public String getHeaderText() { //for printing out FCS files
		StringBuffer buff = new StringBuffer();

		//print standard fields (these get $ at beginning) ... skipped gating params (GnR, etc.) and other ones with "n"
		String[] reqd = { "BEGINANALYSIS" , "BEGINDATA" , "BEGINSTEXT" , "BYTEORD" , "DATATYPE" , "ENDANALYSIS" , "ENDDATA" ,
						"ENDSTEXT" , "MODE" , "NEXTDATA" , "PAR" , "TOT" , "ABRT" , "BTIM" , "CELLS" , "COM" , "COMP" , "CSMODE" ,
						"CSVBITS" , "CYT" , "CYTSN" , "DATE" , "ETIM" , "EXP" , "FIL" , "GATE" , "GATING" , "INST" , "LOST" , 
						"OP" , "PROJ" , "SMNO" , "SRC" , "SYS" , "TIMESTEP" , "TR" , "UNICODE" };

		Set keys = text.keySet();
		for(int i = 0; i < reqd.length; i++) {
			if(keys.contains(reqd[i])) {
				buff.append("\\$"+reqd[i]);
				buff.append("\\"+(String)text.get(reqd[i]));
				keys.remove(reqd[i]);
			}
		}
		for(int i = 0; i < params.length; i++) {
			buff.append("\\$P"+(i+1)+"B\\"+(params[i].len*8));
			keys.remove("P"+(i+1)+"B");
			buff.append("\\$P"+(i+1)+"R\\"+(params[i].range));
			keys.remove("P"+(i+1)+"R");
			buff.append("\\$P"+(i+1)+"E\\"+(params[i].amp1)+","+((int)params[i].amp2));
			keys.remove("P"+(i+1)+"E");
			if(annotations.containsKey("P"+(i+1)))
				buff.append("\\$P"+(i+1)+"S\\"+(String)annotations.get("P"+(i+1)));
			else if(keys.contains("P"+(i+1)+"S"))
				buff.append("\\$P"+(i+1)+"S\\"+(String)text.get("P"+(i+1)+"S"));
			keys.remove("P"+(i+1)+"S");
			annotations.remove("P"+(i+1));
			if(keys.contains("P"+(i+1)+"N"))
				buff.append("\\$P"+(i+1)+"N\\"+(String)text.get("P"+(i+1)+"N"));
			keys.remove("P"+(i+1)+"N");
			if(keys.contains("P"+(i+1)+"DISPLAY"))
				buff.append("\\P"+(i+1)+"DISPLAY\\"+(String)text.get("P"+(i+1)+"DISPLAY"));
			keys.remove("P"+(i+1)+"DISPLAY");
		}
		for (Enumeration e = annotations.keys() ; e.hasMoreElements() ;) {
			String k = (String)e.nextElement();
			buff.append("\\A$"+k+"\\"+(String)annotations.get(k));
			keys.remove(k);
		}
		for(Iterator i = keys.iterator(); i.hasNext(); ) {
			String key = (String)i.next();
			buff.append("\\"+key+"\\");
			buff.append((String)text.get(key));
		}
		buff.append("\\");

		return buff.toString();
	}*/

	public void setAnnotation(String key, Object value) {
		annotations.put(key, value);
	}

	public Object getAnnotation(String key) {
		if(annotations.containsKey(key))
			return annotations.get(key);
		else
			return "";
	}

	public Set getAnnotations() {
		return annotations.keySet();
	}

	/*public int compareTo(Object o) {
		if(!(o instanceof FACSData))
			throw new ClassCastException();
		return filename.compareTo(((FACSData)o).filename);
	}*/

	public String toString() {
		return filename;
	}
}