//FParameter class (for FACS parameters)
package FACS;

import java.nio.ByteBuffer;
import java.nio.ByteOrder;

public class FParameter implements java.io.Serializable {
	private static final long serialVersionUID = 1L;
	public int len, range, amp1;
	public int int_range; //internal store of range for reading ('range' will be chaned to be uniform across parameters)
    public double amp2;
    public boolean log;
    public String lname, sname;
	public boolean isLog;
    
    public FParameter(int len, int range, int amp1, double amp2, String lname, String sname) {
        this.len = len;
        this.range = range;
        int_range = range;
        this.amp1 = amp1;
		if(amp1 == 0)
			isLog = false;
		else
			isLog = true;
        this.amp2 = amp2;
        this.lname = lname;
        this.sname = sname;
        
        if(lname.trim().equals(""))
            this.lname = sname;
        
        if(amp1 == 0 && amp2 == 0)
            log = false;
        else
            log = true;
    }

	public int read(ByteBuffer bb, int byteorder, int datatype) throws ReadException {
        ByteOrder order;
        if(byteorder == 1) {//4,3,2,1
            order = ByteOrder.BIG_ENDIAN;
        } else if(byteorder == 2) { //1,2,3,4
            order = ByteOrder.LITTLE_ENDIAN;
        } else {
            throw new ReadException("3,4,1,2 Byte Order not Supported");
        }
        
        if(datatype == 1) { //integers
            if(len == 2) { //short
                short a = bb.order(order).getShort();
				int b = (new Short(a).intValue()) % range;
				if(b < 0)
					b += range;
                return b*range/int_range;
            } else if(len == 4) { //int
                return bb.order(order).getInt()*(range/int_range);
            }
        } else if(datatype == 2) { //float
            if(len == 4) {
                return (int)bb.order(order).getFloat()*(range/int_range);
            }
        } else if(datatype == 3) { //double
            if(len == 8) {
                return (int)bb.order(order).getDouble()*(range/int_range);
            }
        } else { //ASCII
			byte[] b = new byte[len];
			bb.get(b);
            return new Integer(new String(b)).intValue()*(range/int_range);
        }
        throw new ReadException("Unreadable data type");
    }
}