//////////////
//Thread to do gating

package FACS;

import java.beans.XMLDecoder;
import java.beans.XMLEncoder;
import java.io.*;
import java.util.Vector;
import java.util.Hashtable;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import java.awt.Shape;
import java.awt.Polygon;
import java.awt.geom.*;
import java.nio.channels.FileChannel;
import java.nio.MappedByteBuffer;
import java.nio.ByteBuffer;
import FACS.FParameter;
import FACS.Plot.Region;
import FACS.ReadException;

public class GaterBean implements Runnable, Serializable {
	public static final int NONE = 0;
	public static final int ELLIPSE = 1;
	public static final int RECT = 2;
	public static final int POLY = 3;
	public static final int QUAD = 4;
	public static final int RANGE = 5;

	public static final int BINS = 256;

	private Vector pops;
	private String dir, filedir; //filedir is directory that contains actual FCS files (for dup experiments)
	private StringBuffer out;

	private boolean started;
    private boolean running;
	private double counter;
	private int time;

	public GaterBean() {
		pops = null;
		dir = ""; filedir = "";
		started = false;
		running = false;
		counter = 0;
		time = 0;
		out = new StringBuffer();
	}

	public void setDir(String ROOT, String dir) {
		this.dir = ROOT+dir;
		filedir = ROOT+XMLFiles.getFileLink(ROOT,dir);
	}

	public void setPops(Vector pops) {
		this.pops = pops;
	}

	public synchronized int getPercent() {
		if(pops.size() > 0)
			return (int)Math.rint(counter*100)/pops.size();
		else
			return 100;
    }

	public synchronized int getTimeLeft() {
		if(getPercent() != 0)
			return (int)(100*time/getPercent())-time;
		else
			return -1;
	}

	public synchronized int getTimeElapsed() {
		return time;
	}

    public synchronized boolean isStarted() {
        return started;
    }

    public synchronized boolean isCompleted() {
        return getPercent() == 100;
    }

    public synchronized boolean isRunning() {
        return running;
    }

    public synchronized void setRunning(boolean running) {
        this.running = running;
        if (running)
            started = true;
    }

    public synchronized String getResult() {
        //if (isCompleted())
            return out.toString();
        //else
        //    return null;
    }

	public Vector readRegions() {
		//read in gates (regions.xml)
		File xml = new File(dir,"regions.xml");
		Vector regions = new Vector();
		if(xml.exists()) {
			try {
				XMLDecoder d = new XMLDecoder(new BufferedInputStream(new FileInputStream(xml)));
				while(true) {
					try {
						Region r = (Region)d.readObject();
						regions.add(r);
					} catch(ArrayIndexOutOfBoundsException aioobe) { //file is read completely
						d.close();
						break;
					}
				}
			} catch (FileNotFoundException fnfe) {
				out.append(fnfe);
			}
		}
		return regions;
	}

	public void run() {
		setRunning(true);
		Vector regions = readRegions();
		double[][] c = getComp();
		Vector stats = new Vector();
		for(int i = 0; i < pops.size(); i++) { //for each population
			Population pop = (Population)pops.elementAt(i);
			FileChannel fc; MappedByteBuffer mbb; ByteBuffer bb;
			for(int j = 0; j < pop.getFiles().size(); j++) { //for each file we need to gate
				long tstart = System.currentTimeMillis();
				String fname = (String)pop.getFiles().elementAt(j);
				boolean isComp = pop.isComp() && (c.length != 0);
				try {
					out.append("<h3>Reading "+fname+"... ");
					//read FCS header stuff...
					File f = new File(filedir,fname);
					FileInputStream fis = new FileInputStream(f);
					byte[] text = null;
					int[] pos = getPositions(fis);
					Hashtable info = getInfo(fis,pos[0],pos[1]);
					if(pos[2] == -1) {
						pos[2] = new Integer(((String)info.get("BEGINDATA")).trim()).intValue();
						pos[3] = new Integer(((String)info.get("ENDDATA")).trim()).intValue();
					}
					int byteorder = getByteOrder(info);
					int datatype = getDataType(info);
					FParameter[] params = getParams(info);
					double[] logscale = new double[params.length]; //for lin->log conversion
					double[] logscale2 = new double[params.length];
					int lenSum = 0;
					for(int p = 1; p <= params.length; p++) {
						if(!params[p-1].isLog && params[p-1].log) { //need to make logarithmic
							logscale[p-1] = (double)params[p-1].range/Math.ceil(Math.log10((double)params[p-1].range));
							logscale2[p-1] = logscale[p-1];
						} else if(params[p-1].isLog) {
							logscale[p-1] = (double)params[p-1].amp1/(double)params[p-1].range;
							logscale2[p-1] = (double)params[p-1].range/Math.ceil((double)params[p-1].amp1); //Math.log10((double)params[p-1].range)
						}
						lenSum += params[p-1].len;
					}
					
					//correct data start/end positions if necessary (we've seen FlowJo output a funky file)
					int calcDataLen = 0;
				    for(int p = 0; p < params.length; p++) {
				    	calcDataLen += params[p].len;
				    }
				    calcDataLen *= Integer.parseInt((String)info.get("TOT"));
				    if((pos[3]+1 - pos[2] < calcDataLen)) {
				    	if((pos[2]+calcDataLen <= f.length())) {
				    		pos[3] = pos[2]+calcDataLen-1;
					    } else {
					    	throw new ReadException("File size does not match expected value!");
					    }
				  	}

					//read & gate data
					Vector gates = getGatesByName(params,regions,pop.getGates(),fname);
					boolean isGlobal = getIsGlobalByName(regions,pop.getGates(),fname);
					Vector shapes = getShapes(gates, params);
					double[] sum = zeroes(params.length); double num = 0d;
					double[][] hist = zeroes(params.length,BINS); //hist[channel][bin] !

					fc = fis.getChannel();
					mbb = fc.map(FileChannel.MapMode.READ_ONLY,pos[2],pos[3]-pos[2]+1);
					bb = mbb.asReadOnlyBuffer();
					int nCells = (pos[3]-pos[2]+1)/lenSum;
					
					//int[][] cells = new int[nCells+1][nParam];
					//int i = 0;
					double[] cellLin = new double[params.length]; //used for statistics (& compensation...)
					double[] cellLog = new double[params.length]; //used for gating

					//new File(dir+"a.txt").delete();
					//RandomAccessFile raf = new RandomAccessFile(dir+fname+pop.getName()+".txt","rw");

					while(bb.hasRemaining()) {
						//cells[i] = new int[nParam];
						try {
							for(int p = 0; p < params.length; p++) {
								if(!params[p].isLog && params[p].log) { //need to make logarithmic
									cellLin[p] = (double)params[p].read(bb, byteorder, datatype);
									if(cellLin[p] <= 0.0) //prevent -infinity logs
										cellLog[p] = 0;
									else if(!isComp) //don't need for comp, and will do this calc afterwards
										cellLog[p] = Math.log10((double)cellLin[p])*logscale[p];
								} else if(params[p].isLog) { //need to linearize
									cellLog[p] = (double)params[p].read(bb, byteorder, datatype);
									cellLin[p] = Math.pow(10d,cellLog[p]*logscale[p]);
								} else { //should stay linear
									cellLin[p] = (double)params[p].read(bb, byteorder, datatype);
									cellLog[p] = cellLin[p];
								}
							}
						} catch (ReadException re) {
							continue;
						}
						
						//do compensation here...
						//double[] cellLinOld = cellLin;
						if(isComp) {
							cellLin = multiply(c,cellLin);
							for(int p = 0; p < params.length; p++) {
								if((!params[p].isLog && params[p].log) || params[p].isLog) {
									if(cellLin[p] <= 0.0) //prevent -infinity logs
										cellLog[p] = 0;
									else
										cellLog[p] = Math.log10(cellLin[p])*logscale2[p];
								} else {
									cellLog[p] = cellLin[p];
								}
							}
						}

						//gating
						if(insideGates(cellLog,gates,shapes)) {
							/*if(isComp)
								raf.writeBytes(cellLinOld[3]+"\t"+cellLin[3]+"\n");*/
							addArrays(sum,cellLin); //for mean
							makeHist(hist,cellLin,cellLog,params);
							num++;
						}
					}
					fis.close(); fis = null; System.gc();

					PopStats ps = new PopStats(fname,pop.getName(),isComp,isGlobal);
					getStats(ps,nCells,params,num,sum,hist);

					/*int v = 2;
					for(int z = 0; z < hist[v].length; z++) {
						raf.writeBytes(getBin(params[v],hist[v].length,z)+"\t"+hist[v][z]+"\n");
					}*/
					//raf.close();

					time += (int)(System.currentTimeMillis()-tstart);
					//out.append("("+(System.currentTimeMillis()-tstart)+")</h3><br>"+ps.getHTML());
					stats.add(ps);
				} catch (IOException ioe) {
					ioe.printStackTrace();
					out.append(ioe);
				}
				counter += 1d/pop.getFiles().size();
			}
		}
		updateStatsXML(stats);
		setRunning(false);
	}

	public void updateStatsXML(Vector stats) { //can't just delete old file, since we may be gating only a subset of total gates...
		Vector pops = XMLFiles.getPopulations(dir,"");
		File xml = new File(dir,"stats.xml");
		Vector total = new Vector();
		if(xml.exists()) {
			try {
				XMLDecoder d = new XMLDecoder(new BufferedInputStream(new FileInputStream(xml)));
				while(true) {
					try {
						PopStats ps = (PopStats)d.readObject();
						int add = -1;
						if(!populationExists(pops,ps))
							add = 0;
						for(int i = 0; i < stats.size(); i++) {
							PopStats psi = (PopStats)stats.elementAt(i);
							if(psi.getFile().equals(ps.getFile()) && psi.getPop().equals(ps.getPop())) {
								if(ps.isComp() == psi.isComp()) { //so we can compare pre- and post-compensation (for comp samples)
									add = i;
									break;
								}
							}
						}
						if(add == -1)
							total.add(ps);
					} catch(ArrayIndexOutOfBoundsException aioobe) { //file is read completely
						d.close();
						break;
					}
				}
			} catch (FileNotFoundException fnfe) {
			}
		}
		total.addAll(stats);

		xml.delete();
		try { //write each region to XML file
			FileOutputStream fos = new FileOutputStream(xml);
			XMLEncoder xenc = new XMLEncoder(fos);
			xenc.setExceptionListener(new java.beans.ExceptionListener() {
				public void exceptionThrown(Exception exception) {
					exception.printStackTrace();
				}
			});
			for(int i = 0; i < total.size(); i++) {
				xenc.writeObject(total.elementAt(i));
				xenc.flush();
			}
			xenc.close();
			fos.close();
		} catch (Exception e) {
			e.printStackTrace();
		}
	}
	
	/**
	 * 
	 * 
	 * @param populations Vector of all Populations
	 * @param ps PopStats to test
	 * @return True if PopStats's Population and File are contained in populations
	 */
	public boolean populationExists(Vector populations, PopStats ps) {
		for(int i = 0; i < populations.size(); i++) {
			Population p = (Population)populations.elementAt(i);
			if(p.getName().equals(ps.getPop())) {
				if(p.getFiles().contains(ps.getFile()))
					return true;
				return false;
			}
		}
		return false;
	}

	public int[] getPositions(FileInputStream fis) throws IOException {
		byte[] six = new byte[6];
		fis.read(six);
		String s = new String(six);
		if(!s.equals("FCS1.0") && !s.equals("FCS2.0") && !s.equals("FCS3.0"))
			throw new IOException("File Format Unrecognized\n(says "+s+")");
		
		//spaces
		fis.skip(4);
		
		//starts and ends of TEXT, DATA
		int[] out = new int[4];
		byte[] eight = new byte[8];
		fis.read(eight);
		out[0] = getInt(eight);
		fis.read(eight);
		out[1] = getInt(eight);
		fis.read(eight);
		out[2] = getInt(eight);
		fis.read(eight);
		out[3] = getInt(eight);
		//don't care about analysis
		
		if(out[0] == 0 || out[1] == 0)
			throw new IOException("FCS3.0 Long Files Unsupported If TEXT Section is long");
		
		if(out[2] == 0) { //FCS3 long files (data start/end in text)
			out[2] = -1;
			out[3] = -1;
			/*fis.skip(out[0] - (6+4+8+8+8+8));
            
            text = new byte[out[1]-out[0]+1];
            fis.read(text);
            String txt = new String(text);
            
            //need to get data start & end from text...
            String del = txt.substring(0,1); //delimiter
            Matcher non = Pattern.compile("[^A-Za-z]").matcher(del);
            String del1 = del;
            if(non.find())
                del1 = "\\"+del;
            String[] list = txt.split(del1);
            Pattern patBegin = Pattern.compile("BEGINDATA");
            Pattern patEnd = Pattern.compile("ENDDATA");
            for(int i = 0; i < list.length; i++) {
            	if(patBegin.matcher(list[i]).find()) {
            		out[2] = new Integer(list[i+1].trim()).intValue();
            	} else if(patEnd.matcher(list[i]).find()) {
            		out[3] = new Integer(list[i+1].trim()).intValue();
            	}
            }*/
		}

		return out;
	}

	public Hashtable getInfo(FileInputStream fis, int start, int end) throws IOException {
		fis.skip(start - (6+4+8+8+8+8));
		byte[] text = new byte[end-start+1];
		fis.read(text);
		String txt = new String(text);
		return FCSTools.readText(txt);
	}

	public int getByteOrder(Hashtable info) throws IOException {
		int byteorder;
		if(!info.containsKey("BYTEORD"))
			throw new IOException("File Format Incorrect:\nno byte order specified");
		if(info.get("BYTEORD").equals("4,3,2,1"))
			byteorder = 1;
		else if(info.get("BYTEORD").equals("1,2,3,4"))
			byteorder = 2;
		else if(info.get("BYTEORD").equals("3,4,1,2"))
			byteorder = 3;
		else
			throw new IOException("Unknown byte order -- file likely corrupted.");
		return byteorder;
	}

	public int getDataType(Hashtable info) throws IOException {
		int datatype;
		if(!info.containsKey("DATATYPE"))
			throw new IOException("File Format Incorrect:\nno data type specified");
		if(info.get("DATATYPE").equals("I")) //integer
			datatype = 1;
		else if(info.get("DATATYPE").equals("F")) //float
			datatype = 2;
		else if(info.get("DATATYPE").equals("D")) //double
			datatype = 3;
		else if(info.get("DATATYPE").equals("A")) //ASCII
			datatype = 4;
		else
			throw new IOException("File Format Incorrect:\ndata type unrecognized");
		return datatype;
	}

	public FParameter[] getParams(Hashtable info) throws IOException {
		//int byteorder = getByteOrder(info);
		
		//int datatype = getDataType(info);
		
		int mode;
		if(!info.containsKey("MODE"))
			throw new IOException("File Format Incorrect:\nstorage mode not specified");
		if(info.get("MODE").equals("L")) //list
			mode = 1;
		else if(info.get("MODE").equals("C")) //correlated
			mode = 2;
		else if(info.get("MODE").equals("U")) //uncorr.
			mode = 3;
		else
			throw new IOException("File Format Incorrect:\nmode unrecognized");
		if(mode != 1)
			throw new IOException("Only list data supported.");
		
		int nParam;
		if(!info.containsKey("PAR"))
			throw new IOException("File Format Incorrect:\nparameters unspecified");
		Integer I = new Integer((String)info.get("PAR"));
		nParam = I.intValue();
		
		//set up parameters
		FParameter[] params = new FParameter[nParam];
		for(int i = 1; i <= params.length; i++) { //params start at #1
			if(!info.containsKey("P"+i+"B")) //bit length
				throw new IOException("File Format Incorrect:\nparameter "+i+" incomplete");
			I = new Integer((String)info.get("P"+i+"B"));
			int pb = I.intValue();
			if(pb % 8 != 0)
				throw new IOException("Only byte-justified data supported");
			pb = pb/8;
			
			if(!info.containsKey("P"+i+"R")) //range
				throw new IOException("File Format Incorrect:\nparameter "+i+" incomplete");
			I = new Integer((String)info.get("P"+i+"R"));
			int pe = I.intValue();
			
			if(!info.containsKey("P"+i+"E")) //amplification
				info.put("P"+i+"E","0,0");
				//throw new IOException("File Format Incorrect:\nparameter "+i+" incomplete");
			int comma = ((String)info.get("P"+i+"E")).indexOf(',');
			String Spr1 = ((String)info.get("P"+i+"E")).substring(0,comma);
			String Spr2 = ((String)info.get("P"+i+"E")).substring(comma+1);
			Double d = new Double(Spr1); //was Integer for FCS2.0...
			int pr1 = d.intValue();
			d = new Double(Spr2);
			double pr2 = d.doubleValue();
			
			String ps;
			if(info.containsKey("P"+i+"S"))
				ps = (String)info.get("P"+i+"S"); //long name
			else
				ps = "";
			
			String pn;
			
			if(info.containsKey("P"+i+"N"))
				pn = (String)info.get("P"+i+"N"); //short
			else
				pn = "P"+i;
			
			params[i-1] = new FParameter(pb, pe, pr1, pr2, ps, pn);
			if(info.containsKey("P"+i+"DISPLAY")) {
				String islog = (String)info.get("P"+i+"DISPLAY");
				if(islog.equalsIgnoreCase("log")) {
					params[i-1].log = true;
					params[i-1].amp1 = (int)Math.ceil(Math.log((double)params[i-1].range)/Math.log(10.0));
				}
			}
		}
		
		//we're going to normalize all parameter ranges to one value
        int paramRange = 0;
        for(int i = 0; i < params.length; i++) {
        	if(params[i].range > paramRange)
        		paramRange = params[i].range;
        }
        for(int i = 0; i < params.length; i++)
        	params[i].range = paramRange;
		
		return params;
	}
		
	public int getInt(byte[] b) throws NumberFormatException {
		String s = new String(b);
		Integer i = new Integer(s.trim());
		return i.intValue();
	}
		
	/*public Hashtable readText(String txt) { //get key-value pairs out of list; keys have "$" at beginning; all are separated by a delimiter
		String del = txt.substring(0,1); //delimiter
		Matcher non = Pattern.compile("[^A-Za-z]").matcher(del);
		String del1 = del;
		if(non.find())
			del1 = "\\"+del;
		String[] list = txt.split(del1);
		
		Hashtable hash = new Hashtable();
		String key = "";
		for(int i = 0; i < list.length; i++) {
			if(i == 0 && list[i].equals(""))
				i++;
			if(list[i].equals("")) {
				if(i >= list.length)
					break;
				String t = (String)hash.get(key);
				i++;
				hash.put(key,t+del+list[i]);
			} else {
				if(i >= list.length)
					break;
				if(list[i].startsWith("$"))
					key = list[i].substring(1);
				else
					key = list[i];
				i++;
				if(i >= list.length) //just if it recognizes something after the last delimiter
                	break;
				hash.put(key,list[i]);
			}
		}
		return hash;
	}*/

	public Vector getGatesByName(FParameter[] params, Vector regions, Vector names, String fname) { //return gate parameters (double[]) from list of gate names
		Vector out = new Vector();
		for(int i = 0; i < names.size(); i++) {
			String name = (String)names.elementAt(i);
			for(int j = 0; j < regions.size(); j++) {
				Region r = (Region)regions.elementAt(j);
				if(r.getName().equals(name)) {
					//out.add(regions.elementAt(j);
					out.add(r.getVals(fname));
					break;
				} else if(name.startsWith(r.getName()) && isQuad(name)) {
					double[] quad = getQuad(params,name,r.getVals(fname));
					if(quad != null)
						out.add(quad);
				}
			}
		}
		return out;
	}
	
	public boolean isQuad(String name) {
		if(name.endsWith("-UL"))
			return true;
		else if(name.endsWith("-UR"))
			return true;
		else if(name.endsWith("BL"))
			return true;
		else if(name.endsWith("BR"))
			return true;
		return false;
	}
	
	public double[] getQuad(FParameter[] params, String name, double[] region) {
		if((int)region[0] != QUAD)
            return null;
        
		int num = 0;
		if(name.endsWith("-UL"))
			num = 1;
		else if(name.endsWith("-UR"))
			num = 2;
		else if(name.endsWith("-BL"))
			num = 3;
		else if(name.endsWith("-BR"))
			num = 4;
		
		double[] out = new double[7];
		out[0] = RECT;
		out[1] = region[1];
		out[2] = region[2];
        switch(num) {
            case 3: //BL
            	out[3] = 0.0;
            	out[4] = 0.0;
            	out[5] = region[3];
            	out[6] = region[4];
            	return out;
            case 4: //BR
            	out[3] = region[3];
            	out[4] = 0.0;
            	out[5] = params[(int)region[1]].range-region[3];
            	out[6] = region[4];
            	return out;
            case 1: //UL
            	out[3] = 0.0;
            	out[4] = region[4];
            	out[5] = region[3];
            	out[6] = params[(int)region[2]].range-region[4];
            	return out;
            case 2: //UR
            	out[3] = region[3];
            	out[4] = region[4];
            	out[5] = params[(int)region[1]].range-region[3];
            	out[6] = params[(int)region[2]].range-region[4];
            	return out;
        }
        return null;
	}

	public boolean getIsGlobalByName(Vector regions, Vector names, String fname) { //are all gates global?
		boolean isGlobal = true;
		for(int i = 0; i < names.size(); i++) {
			for(int j = 0; j < regions.size(); j++) {
				if(((Region)regions.elementAt(j)).getName().equals((String)names.elementAt(i))) {
					isGlobal = (isGlobal && ((Region)regions.elementAt(j)).isGlobal(fname));
				}
			}
		}
		return isGlobal;
	}

	public Vector getShapes(Vector gates, FParameter[] params) { //turn gate parameters into shapes
		Vector out = new Vector();
		double[] gate; Shape s;
		for(int i = 0; i < gates.size(); i++) {
			gate = (double[])gates.elementAt(i);
			s = null;
			if((int)gate[0] == RECT) {
				s = new Rectangle2D.Double(gate[3],gate[4],gate[5],gate[6]);
			} else if((int)gate[0] == ELLIPSE) {
				Ellipse2D.Double el = new Ellipse2D.Double(0,-gate[6]/2,gate[5],gate[6]);
				AffineTransform atEl = AffineTransform.getTranslateInstance(gate[3],gate[4]+gate[6]/2);
				atEl.rotate(-gate[7]);
				s = atEl.createTransformedShape(el);
			} else if((int)gate[0] == RANGE) {
				double theta = Math.atan((gate[6]-gate[4])/(gate[5]-gate[3]));
				if((theta < Math.PI/5 && theta > -Math.PI/5) || gate[2] == -1)
					s = new Rectangle2D.Double(gate[3],0.0,gate[5]-gate[3],params[(int)gate[1]].range); //horiz
				else
					s = new Rectangle2D.Double(0.0,gate[6],params[(int)gate[2]].range,gate[4]-gate[6]);
			} else if((int)gate[0] == POLY) {
				int[] xp = new int[(gate.length-3)/2];
				int[] yp = new int[(gate.length-3)/2];
				int n = 0;
				for(int p = 3; p < gate.length; p++) {
					xp[n] = (int)gate[p];
					p++;
					yp[n] = (int)gate[p];
					n++;
				}
				s = new Polygon(xp,yp,(gate.length-3)/2);
			}
			out.add(s);
		}
		return out;
	}

	public boolean insideGates(double[] cell, Vector gates, Vector shapes) {
		Shape s; double[] gparams;
		for(int i = 0; i < gates.size(); i++) {
			s = (Shape)shapes.elementAt(i);
			gparams = (double[])gates.elementAt(i);
			if(s == null)
				continue;
			if(!s.contains(cell[(int)gparams[1]],cell[(int)gparams[2]])) {
				return false;
			}
		}
		return true;
	}

	public double[] zeroes(int len) {
		double[] out = new double[len];
		for(int i = 0; i < out.length; i++) {
			out[i] = 0d;
		}
		return out;
	}

	public double[][] zeroes(int len1, int len2) {
		double[][] out = new double[len1][len2];
		for(int i = 0; i < len1; i++) {
			out[i] = zeroes(len2);
		}
		return out;
	}

	public double[] addArrays(double[] one, double[] two) { //this function increments the first argument!!! (be careful)
		for(int i = 0; i < one.length; i++) {
			one[i] += two[i];
		}
		return one;
	}

	public double[][] makeHist(double[][] hist, double[] cellLin, double[] cellLog, FParameter[] params) { //put this cell's values into the histogram
		int bins = hist[0].length;
		int pos;
		for(int i = 0; i < params.length; i++) {
			if(params[i].log)
				pos = (int)(cellLog[i]*bins/params[i].range);
			else
				pos = (int)(cellLin[i]*bins/params[i].range);
			try {
				hist[i][pos]++;
			} catch(ArrayIndexOutOfBoundsException aioobe) { } //just in case... so we don't just die
		}
		return hist;
	}

	/*public int getLinRange(FParameter param) {
		if(!param.isLog && param.log) { //need to make logarithmic
			return param.range;
		} else if(param.isLog) { //need to linearize
			return (int)Math.pow(10d,param.amp1);
		} else { //should stay linear
			return param.range;
		}
	}*/

	//[maximal %error on the median (i.e. percent difference between consecutive bins), given 6-log range of values
	// = 10^(amp1/bins)-1 = 10^(6/bins): 20 => 100%, 150 => 10%, 250 => 5.5%, 500 => 2.8%, 1000 => 1.4%
	public static double getBin(FParameter param, int bins, int i) {
		if(!param.log)
			return ((double)i+0.5)*((double)param.range/(double)bins);
		else
			return Math.pow(10d,((double)i+0.5)*((double)param.amp1/(double)bins));
	}

	public static double getDelta(FParameter param, int bins, int i) {
		if(!param.log)
			return (double)param.range/bins;
		else
			return Math.pow(10d,((double)i)*((double)param.amp1/(double)bins)) - Math.pow(10d,((double)i-1)*((double)param.amp1/(double)bins));
	}

	public static PopStats getStats(PopStats ps, int nCells, FParameter[] params, double num, double[] sum, double[][] hist) {
		ps.setPct(num*100/(double)nCells);
		ps.setNum(num);
		ps.setSize(params.length);
		//output is double[channel][statistic]
		//where statistic: 0=mean, 1=median, 2=cv (=sqrt(variance)/mean*100)
		//double[][] stats = new double[sum.length][3];
		double px, running; //running sum
		double x; //double xmult; //actual number corresponding to histogram bin
		boolean medfound; //found median
		for(int i = 0; i < params.length; i++) { //for each channel
			ps.setMean(i,sum[i] / num); //mean
			running = 0;
			//xmult = getLinRange(params[i])/hist[i].length;
			medfound = false;
			double cv = 0;
			for(int j = 0; j < hist[i].length; j++) { //for each bin
				px = hist[i][j] / num;
				running += px;
				x = getBin(params[i],hist[i].length,j); //(double)(j+1)*xmult; //turns out that j*xmult is the LOWER BOUND of the bin
				if(running > 0.5 && !medfound) { //find median... we'll interpolate from the upper bound (could also average both directions)
					ps.setMedian(i,(0.5-running)/px*getDelta(params[i],hist[i].length,j)+x);
					medfound = true;
				}
				cv += Math.pow(x-ps.getMean()[i],2d)*px; //our "x" should represent halfway in between
			}
			ps.setCV(i,Math.sqrt(cv) / ps.getMean()[i] * 100d);
		}
		return ps;
	}
	
	public double[][] getComp() {
		File xml = new File(dir,"comp.xml");
		double[][] c = new double[0][0];
		if(xml.exists()) { //don't bother loading of comp == false
			try {
				XMLDecoder d = new XMLDecoder(new BufferedInputStream(new FileInputStream(xml)));
				try {
					c = (double[][])d.readObject();
				} finally { //file is read completely
					d.close();
				}
			} catch (IOException ioe) {
				ioe.printStackTrace();
			}
		}
		return c;
	}

	public double[] multiply(double[][] comp, double[] cell) { //perform matrix*vector multiplication
		if(comp[0].length != cell.length) //matrix wrong size
			return cell;
		double[] out = new double[cell.length];
		for(int i = 0; i < cell.length; i++) {
			out[i] = dotProd(comp[i],cell);
		}
		return out;
	}

	public double dotProd(double[] one, double[] two) { //dot product of two vectors (as part of matrix mult)
		double out = 0;
		for(int i = 0; i < one.length; i++) {
			out += one[i]*two[i];
		}
		return out;
	}
}