//////////////////////////
// Class for displaying FACS Plots, modified for applets
//
// Copyright 2004 Mark Hammer. All rights reserved.
//
//////////////////////////

package FACS.Plot.PlotApplet;

import FACS.FACSData;
import FACS.FParameter;
import FACS.Plot.Region;
import FACS.Plot.RVector;
import javax.swing.*;
import javax.swing.event.*;
import javax.swing.table.*;
import java.awt.*;
import java.awt.event.*;
import java.awt.geom.*;
import java.awt.image.*;
import java.util.*;
import java.util.regex.*;

public class Plot extends JComponent implements MouseListener, ActionListener, ItemListener, MouseMotionListener {
    //note: even though can have multiple data sets, parameter array will be taken from FIRST one

	static final int NONE = 0;
    static final int ELLIPSE = 1;
    static final int RECT = 2;
    static final int POLY = 3;
    static final int QUAD = 4;
    static final int RANGE = 5;
    
    static Dimension preferredSize = new Dimension(340,355);
    
    boolean draw;
    int[][][] data;
    int param1, param2;
    FParameter[][] params;
    FACSData[] fData;
    int[] dataIndices;
    /*PlotPanel pp;*/
	int region;
    PlotApplet facs;
    Vector gates; //gates for the DATA

	boolean vectorOutput; //output to vector graphics -> repaint all points directly
    
    Color[] colors;
    
    int mode;
    final int DOTPLOT = 1;
    final int HIST = 2;
    final int SM_HIST = 3;
    final int DENSITY = 4;
    final int CONTOUR = 5;
    boolean hSecondMax; //histogram normalized while ignoring 5 last bins
    boolean smooth; //option for density plot
    boolean cShowOutliers; //show outliers in contour plot
    int nLevels;
    int h2DSmooth; //resolution scale-down factor for 2D histograms
    int hSmooth2; //averaging smoothening factor
    int old2D, old1D; //user's last selections
    
    JCheckBoxMenuItem cbSmooth, cbDSmooth, cb2Max;
    JCheckBoxMenuItem cbOutliers;
    JRadioButtonMenuItem rbDensity, rbDot, rbContour;
    
    boolean log1, log2;
    float max1, max2;
    float min1, min2;
    
    int resX; //resolution at which pic is actually drawn (before scale)
    int resY;
    
    final int title = 15; //height of title area
    final int axisBorder = 20; //borders on sides
    final int frameBorder = 15; //length of tick mark + labels
	final int tickSize = 5; //actual length of tick marks

	final int gateDragBoxSize = 5; //size of those drag boxes on the gates... (1/2 width i think)
    
    static int xTicks = 20; //axis tick marks
    static int yTicks = 20;
    
    BufferedImage bi;
    
    double w, h;
    
    boolean selected;
    
    //region stuff
    int rStartX, rStartY;
    int regionEditing, paramEdit;
    RVector regions; //all gates in the project
    Vector rDisplayed;
    //Vector percents;
	Vector rLocked;

	String defaultRegionName;
    
    public Plot(PlotApplet facs) { //PlotPanel pp, FACSManip facs
        /*this.pp = pp;*/
        this.facs = facs;
        gates = new Vector();
        regions = new RVector();
        //percents = null;
        rDisplayed = null;
        bi = new BufferedImage(300,300,BufferedImage.TYPE_3BYTE_BGR);
        draw = false;
        selected = false;
		vectorOutput = false;

		defaultRegionName = null;
        
        if(rDisplayed == null) {
            rDisplayed = new Vector();
            for(int i = 0; i < facs.regions.size(); i++) {
            	Region r = (Region)facs.regions.elementAt(i);
            	if(r.getName().equals(defaultRegionName) || !isCompGate(r.getName()))
            		rDisplayed.add(new Boolean(true));
            	else
            		rDisplayed.add(new Boolean(false));
            }
        }
		
        regions = facs.regions; //just for data choosing purposes
        
        old2D = DOTPLOT;
        old1D = HIST;
        
        hSecondMax = true;
        smooth = false;
        cShowOutliers = true;
        h2DSmooth = 4;
        hSmooth2 = 4;
        nLevels = 3;
        
        cbSmooth = new JCheckBoxMenuItem("Smooth");
        cbSmooth.setActionCommand("smooth");
        cbSmooth.addItemListener(this);
        
        cbDSmooth = new JCheckBoxMenuItem("Smooth");
        cbDSmooth.setActionCommand("dsmooth");
        cbDSmooth.addItemListener(this);
        
        cb2Max = new JCheckBoxMenuItem("Normalize to Second Highest");
        cb2Max.setActionCommand("second");
        cb2Max.setSelected(true);
        cb2Max.addItemListener(this);
        
        cbOutliers = new JCheckBoxMenuItem("Show Outliers");
        cbOutliers.setActionCommand("outliers");
        cbOutliers.setSelected(true);
        cbOutliers.setEnabled(false);
        cbOutliers.addItemListener(this);
        
        ButtonGroup b2D = new ButtonGroup();
        
        rbDot = new JRadioButtonMenuItem("Dot Plot");
        rbDot.setActionCommand("dot");
        rbDot.setSelected(true);
        rbDot.addActionListener(this);
        b2D.add(rbDot);
        
        rbDensity = new JRadioButtonMenuItem("Density Plot");
        rbDensity.setActionCommand("density");
        rbDensity.addActionListener(this);
        b2D.add(rbDensity);
        
        rbContour = new JRadioButtonMenuItem("Contour Plot");
        rbContour.setActionCommand("contour");
        rbContour.addActionListener(this);
        b2D.add(rbContour);
        
        mode = DOTPLOT;
        
        param1 = -1;
        param2 = -1;
        
        rStartX = -1;
        rStartY = -1;
        regionEditing = -1;
        paramEdit = 0;
        
        addMouseListener(this);
        addMouseMotionListener(this);
        
        Action delete = new AbstractAction() {
            public void actionPerformed(ActionEvent e) {
                if(regionEditing != -1) {
                    delete(regionEditing);
                }
            }
        };
        
        getInputMap().put(KeyStroke.getKeyStroke("DELETE"), "delete");
        getInputMap().put(KeyStroke.getKeyStroke("BACK_SPACE"), "delete");
        getActionMap().put("delete",delete);
    }
    
    /**
     * Returns true if it's a gate drawn for compensation "p{i}N" or "p{i}P" where {i} is an integer (or no i for global gate)
     * @param name
     * @return
     */
    public boolean isCompGate(String name) {
    	Matcher m = Pattern.compile("p\\d{0,1}[PN]").matcher(name);
    	if(m.find())
    		return true;
    	return false;
    }

	public void setRegion(String region) {
		for(int i = 0; i < regions.size(); i++) {
			Region r = (Region)regions.elementAt(i);
			if(r.getName().equals(region)) {
				rDisplayed.setElementAt(new Boolean(true),i);
				rLocked.setElementAt(new Boolean(false),i);
			} else {
				rDisplayed.setElementAt(new Boolean(false),i);
			}
		}
		defaultRegionName = region; //in case a region hasn't already been drawn...
	}
    
    public void delete(int regionEditing) {
        regions.removeElementAt(regionEditing);
        //rDisplayed.removeElementAt(regionEditing); //is done in updatePlot
        //percents.removeElementAt(regionEditing);
        facs.regions.removeElementAt(regionEditing);
        facs.updatePlots(regionEditing, false);
        this.regionEditing = -1;
        repaint();
    }

	public RVector translate(RVector regions) {
		RVector out = new RVector();
        for(int i = 0; i < regions.size(); i++) {
			Region old = (Region)regions.elementAt(i);
			out.add(translateRegion(old, i));
		}
        return out;
	}
    
    public Region translateRegion(Region old, int i) { //translates FACS numbers into pixel numbers
		Region out = (Region)old.clone();
		for (Enumeration e = old.getValues().keys() ; e.hasMoreElements() ;) {
			String key = (String)e.nextElement();
			double[] facsRegion = old.getVals(key);
			double[] xyRegion = new double[facsRegion.length];
			if(((Boolean)rDisplayed.elementAt(i)).booleanValue())
				xyRegion[0] = facsRegion[0];
			else
				xyRegion[0] = -facsRegion[0];
			xyRegion[1] = facsRegion[1];
			xyRegion[2] = facsRegion[2];
			if((int)facsRegion[0] == RECT) {
				xyRegion[3] = iScaleToParam(facsRegion[3],(int)facsRegion[1],false);
				xyRegion[4] = params[0][(int)facsRegion[2]].range-facsRegion[4];
				xyRegion[4] = iScaleToParam(xyRegion[4],(int)facsRegion[2],true);
				xyRegion[3] = xyRegion[3]+(axisBorder+frameBorder);
				xyRegion[4] = xyRegion[4]+frameBorder+title;
				xyRegion[5] = iScaleToParam(facsRegion[5],(int)facsRegion[1],false);
				xyRegion[6] = iScaleToParam(facsRegion[6],(int)facsRegion[2],true);
				xyRegion[4] -= xyRegion[6]; //need to switch corners b/c upside-down (& rectangle is retarded)
			} else if((int)facsRegion[0] == ELLIPSE) {
				xyRegion[3] = iScaleToParam(facsRegion[3],(int)facsRegion[1],false);
				xyRegion[4] = params[0][(int)facsRegion[2]].range-facsRegion[4];
				xyRegion[4] = iScaleToParam(xyRegion[4],(int)facsRegion[2],true);
				xyRegion[3] = xyRegion[3]+(axisBorder+frameBorder);
				xyRegion[4] = xyRegion[4]+frameBorder+title;
				xyRegion[5] = iScaleToParam(facsRegion[5],(int)facsRegion[1],false);
				xyRegion[6] = iScaleToParam(facsRegion[6],(int)facsRegion[2],true);
				xyRegion[4] -= xyRegion[6];
				xyRegion[7] = facsRegion[7];
			} else if((int)facsRegion[0] == RANGE) {
				xyRegion[3] = iScaleToParam(facsRegion[3],(int)facsRegion[1],false);
				xyRegion[4] = params[0][Math.abs((int)facsRegion[2])].range-facsRegion[4];
				xyRegion[4] = iScaleToParam(xyRegion[4],Math.abs((int)facsRegion[2]),true);
				xyRegion[3] = xyRegion[3]+(axisBorder+frameBorder);
				xyRegion[4] = xyRegion[4]+frameBorder+title;
				
				xyRegion[5] = iScaleToParam(facsRegion[5],(int)facsRegion[1],false);
				xyRegion[5] = xyRegion[5]+(axisBorder+frameBorder);
				xyRegion[6] = params[0][Math.abs((int)facsRegion[2])].range-facsRegion[6];
				xyRegion[6] = iScaleToParam(xyRegion[6],Math.abs((int)facsRegion[2]),true);
				xyRegion[6] = xyRegion[6]+frameBorder+title;
			} else if((int)facsRegion[0] == POLY) {
				for(int p = 3; p < facsRegion.length; p++) {
					xyRegion[p] = iScaleToParam(facsRegion[p],(int)facsRegion[1],false);
					xyRegion[p] = xyRegion[p]+(axisBorder+frameBorder);
					p++;
					xyRegion[p] = params[0][(int)facsRegion[2]].range-facsRegion[p];
					xyRegion[p] = iScaleToParam(xyRegion[p],(int)facsRegion[2],true);
					xyRegion[p] = xyRegion[p]+frameBorder+title;
				}
			} else if((int)facsRegion[0] == QUAD) {
				xyRegion[3] = iScaleToParam(facsRegion[3],(int)facsRegion[1],false);
				xyRegion[4] = params[0][(int)facsRegion[2]].range-facsRegion[4];
				xyRegion[4] = iScaleToParam(xyRegion[4],(int)facsRegion[2],true);
				xyRegion[3] = xyRegion[3]+(axisBorder+frameBorder);
				xyRegion[4] = xyRegion[4]+frameBorder+title;
			}

			out.setVals(key,xyRegion);
		}
		return out;
    }
    
    public double iScaleToParam(double in, int param, boolean isY) {
        int max;
        if(!isY)
            max = getWidth();
        else
            max = getHeight()-title;
        double out = in * (max-axisBorder-2*frameBorder);
        out /= params[0][param].range;
        return out;
    }

	public RVector unTranslateAll(RVector regions) {
		RVector out = new RVector();
        for(int i = 0; i < regions.size(); i++) {
			Region old = (Region)regions.elementAt(i);
			out.add(unTranslate(old));
		}
        return out;
	}
    
    public Region unTranslate(Region region) {
        //gives region in terms of FACS parameter numbers
        //basically, just translate and scale coords
		Region outregion = (Region)region.clone();
		for (Enumeration e = region.getValues().keys() ; e.hasMoreElements() ;) {
			String key = (String)e.nextElement();

			double[] XYregion = region.getVals(key);
			double[] out = new double[XYregion.length];
			
			out[0] = Math.abs(XYregion[0]);
			out[1] = XYregion[1];
			out[2] = XYregion[2];
			if((int)out[0] == RECT) {
				out[3] = XYregion[3]-(axisBorder+frameBorder);
				out[3] = scaleToParam(out[3],param1,false);
				out[4] = XYregion[4]-frameBorder-title;
				out[4] = params[0][param2].range-scaleToParam(out[4],param2,true); //y is inverted...
				out[5] = scaleToParam(XYregion[5],param1,false);
				out[6] = scaleToParam(XYregion[6],param2,true);
				out[4] -= out[6]; //need to switch corners b/c upside-down (& rectangle is retarded)
			} else if((int)out[0] == ELLIPSE) {
				out[3] = XYregion[3]-(axisBorder+frameBorder);
				out[3] = scaleToParam(out[3],param1,false);
				out[4] = XYregion[4]-frameBorder-title;
				out[4] = params[0][param2].range-scaleToParam(out[4],param2,true); //y is inverted...
				out[5] = scaleToParam(XYregion[5],param1,false);
				out[6] = scaleToParam(XYregion[6],param2,true);
				out[4] -= out[6];
				out[7] = XYregion[7];
			} else if((int)out[0] == RANGE) {
				out[3] = XYregion[3]-(axisBorder+frameBorder);
				out[3] = scaleToParam(out[3],param1,false);
				out[4] = XYregion[4]-frameBorder-title;
				out[4] = params[0][Math.abs(param2)].range-scaleToParam(out[4],Math.abs(param2),true);
				double theta = Math.atan((XYregion[6]-XYregion[4])/(XYregion[5]-XYregion[3]));
				if((theta < Math.PI/5 && theta > -Math.PI/5) || XYregion[2] < 0) {
					out[6] = out[4];
					out[5] = XYregion[5]-(axisBorder+frameBorder);
					out[5] = scaleToParam(out[5],param1,false);
				} else {
					out[5] = out[3];
					out[6] = XYregion[6]-frameBorder-title;
					out[6] = params[0][param2].range-scaleToParam(out[6],param2,true);
				}
			} else if((int)out[0] == POLY) {
				for(int p = 3; p < XYregion.length; p++) {
					out[p] = XYregion[p]-(axisBorder+frameBorder);
					out[p] = scaleToParam(out[p],param1,false);
					p++;
					out[p] = XYregion[p]-frameBorder-title;
					out[p] = params[0][param2].range-scaleToParam(out[p],param2,true);
				}
			} else if((int)out[0] == QUAD) {
				out[3] = XYregion[3]-(axisBorder+frameBorder);
				out[3] = scaleToParam(out[3],param1,false);
				out[4] = XYregion[4]-frameBorder-title;
				out[4] = params[0][param2].range-scaleToParam(out[4],param2,true);
			}
			outregion.setVals(key,out);
		}
		return outregion;
    }
    
    public double scaleToParam(double in, int param, boolean isY) {
        int max;
        if(!isY)
            max = getWidth();
        else
            max = getHeight()-title;
        double out = in / (max-axisBorder-2*frameBorder);
        out *= params[0][param].range;
        return out;
    }
    
    public double[] getCtr(double[] region, int num) {
        double[] ctr = new double[2];
        
        if(Math.abs((int)region[0]) == RECT) {
            ctr[0] = region[3]+region[5]/2;
            ctr[1] = region[4]+region[6]/2;
        } else if(Math.abs((int)region[0]) == POLY) {
            ctr[0] = 0; ctr[1] = 0;
            for(int i = 3; i < region.length; i++) {
                ctr[0] += region[i];
                i++;
                ctr[1] += region[i];
            }
            ctr[0] /= (region.length-3)/2;
            ctr[1] /= (region.length-3)/2;
        } else if(Math.abs((int)region[0]) == ELLIPSE) {
            Ellipse2D.Double el = new Ellipse2D.Double(0,-region[6]/2,region[5],region[6]);
            AffineTransform atEl = AffineTransform.getTranslateInstance(region[3],region[4]+region[6]/2);
            atEl.rotate(region[7]);
            Shape s = atEl.createTransformedShape(el);
            Rectangle2D r = s.getBounds2D();
            ctr[0] = r.getX()+r.getWidth()/2;
            ctr[1] = r.getY()+r.getHeight()/2;
        } else if(Math.abs((int)region[0]) == QUAD) {
            switch(num) {
                case 1: //UL
                    ctr[0] = region[3]/2;
                    ctr[1] = region[4]/2;
                    break;
                case 2: //UR
                    ctr[0] = region[3]+(getWidth()-region[3])/2;
                    ctr[1] = region[4]/2;
                    break;
                case 3: //BL
                    ctr[0] = region[3]/2;
                    ctr[1] = region[4]+(getHeight()-region[4])/2;
                    break;
                case 4: //BR
                    ctr[0] = region[3]+(getWidth()-region[3])/2;
                    ctr[1] = region[4]+(getHeight()-region[4])/2;
                    break;
            }
        } else if(Math.abs((int)region[0]) == RANGE) {
            double theta = Math.atan((region[6]-region[4])/(region[5]-region[3]));
            if((theta < Math.PI/5 && theta > -Math.PI/5) || region[2] < 0) {
                ctr[0] = (region[3]+region[5])/2;
                ctr[1] = region[4];
            } else {
                ctr[0] = region[3];
                ctr[1] = (region[4]+region[6])/2;
            }
        }
        return ctr;
    }
    
    public void setPlot(int[] indices, FACSData[] fData, int param1, int param2, int mode) { //FParameter[] params, int[][] data,
        dataIndices = indices;
        this.fData = fData;
        if(fData.length == 0) {
            draw = false;
            return;
        }

		if(rLocked == null) {
            rLocked = new Vector();
			Region r;
            for(int i = 0; i < facs.regions.size(); i++) {
				r = (Region)facs.regions.elementAt(i);
				if(r.isGlobal(fData[0].filename)) //global gates locked by default
	                rLocked.add(new Boolean(true));
				else
					rLocked.add(new Boolean(false));
            }
        }
        
        //param1 is x-axis, param2 is y-axis
        draw = true;
        //data = new int[fData.length][][];
        params = new FParameter[fData.length][];
        for(int i = 0; i < fData.length; i++) {
            //data[i] = fData[i].data;
            params[i] = fData[i].params;
        }
        
        if(rDisplayed == null) {
            rDisplayed = new Vector();
            for(int i = 0; i < facs.regions.size(); i++) {
                rDisplayed.add(new Boolean(true));
            }
        }
        regions = translate(facs.regions);
        
        /*if(percents == null) {
            percents = new Vector();
            for(int i = 0; i < facs.regions.size(); i++) {
                percents.add(generatePctLabel(regions,i));
            }
        }*/
        
        if(param1 == -1 && param2 == -1) {
            param1 = 0;
            param2 = 1;
            for(int i = 0; i < params[0].length; i++) { //make it actually the FSC, SSC
                String n = params[0][i].lname;
                Matcher fsc = Pattern.compile("fsc",Pattern.CASE_INSENSITIVE).matcher(n);
                Matcher ssc = Pattern.compile("ssc",Pattern.CASE_INSENSITIVE).matcher(n);
                Matcher wa = Pattern.compile("[wa]",Pattern.CASE_INSENSITIVE).matcher(n);
                //we don't want FSC-W or FSC-A
                if(fsc.find() && !wa.find())
                    param1 = i;
                else if(ssc.find() && !wa.find())
                    param2 = i;
            }
        }
        
        if(param2 != -1) {
            mode = old2D;
            if(fData.length > 1)
                mode = DOTPLOT;
        } else {
            mode = old1D;
        }
        
        /*colors = new Color[data.length];
        colors[0] = Color.black;
        for(int i = 1; i < data.length; i++) {
            colors[i] = Color.getHSBColor((float)i/(float)data.length,1.0f,1.0f);
        }*/

        setParams(param1,param2,mode);
        
    }
    
    public void setParams(int param1, int param2, int mode) {
        this.param1 = param1;
        this.param2 = param2;
        this.mode = mode;
        
        if(mode == DENSITY || mode == CONTOUR || mode == HIST || mode == SM_HIST) {
            /*resX = (int)params[0][param1].range/h2DSmooth;
            resY = (int)params[0][param2].range/h2DSmooth;*/
			resX = 256;
			resY = 256;
        } else {
            resX = 300;
            resY = 300;
        }
        /*bi = new BufferedImage(resX,resY,BufferedImage.TYPE_4BYTE_ABGR);
        Graphics2D g2 = (Graphics2D)bi.getGraphics();
		paintPlot(g2);*/

		log1 = params[0][param1].log;
		if(param2 != -1)
			log2 = params[0][param2].log;
		else
			log2 = false;

		bi = facs.getPlot(dataIndices,param1,param2,mode);

        repaint();
    }

	/*public void paintPlot(Graphics2D g2) {
		int param1 = this.param1, param2 = this.param2, mode = this.mode;
		g2.setColor(Color.white);
        g2.fillRect(0,0,resX,resY);

        for(int d = 0; d < data.length; d++) { //will only be multiple for HIST and DOTPLOT
            param1 = this.param1;
            param2 = this.param2;
            
            if(params[d].length != params[0].length)
                continue;
            if(!(mode == DOTPLOT || mode == HIST || mode == SM_HIST) && d > 0) //just in case...
                break;
            
            //generate plot points
            int[][] pts = data[d];
            
            log1 = params[0][param1].log;
            if(param2 != -1)
                log2 = params[0][param2].log;
            else
                log2 = false;
            max1 = (float)params[0][param1].range;
            if(param2 != -1) {
                max2 = (float)params[0][param2].range;
            } else {
                pts = genHistogram(data[d],param1);
                max2 = getMax(pts);
                
                param1 = 0; //these are just the ones for this function
                param2 = 1;
            }
            
            int[][] twoHist = null;
            int dMax = 0;
            if(mode == DENSITY || mode == CONTOUR) {
                //if(mode == DENSITY)
                    twoHist = gen2DHist(data[d],resX,1);
                //else
                //    twoHist = gen2DHist(data[d],h2DSmooth,hSmooth2);
                dMax = get2DMax(twoHist);
                dMax /= 1;
                
                if(mode == CONTOUR) {
                    int[] levels = new int[nLevels];
                    for(int i = 1; i <= nLevels; i++) {
                        levels[i-1] = dMax/nLevels*i/2;
                    }
                    
                    pts = contour(twoHist,levels,cShowOutliers);
                } else if(mode == DENSITY) {
                    pts = ones(param1, param2, params[0][param1].range/resX);
                }
                
                param1 = 0;
                param2 = 1;
                
                //max1 = (float)params[0][param1].range/h2DSmooth;
                //resX = (int)max1;
                //max2 = (float)params[0][param2].range/h2DSmooth;
                //resY = (int)max2;
				max1 = resX; max2 = resY;
            }
            
            //actual rendering
            g2.setColor(colors[d]);
            float lastX = 0, lastY = 0;
            for(int i = 0; i < pts.length; i++) {
                float x = pts[i][param1]/max1*(float)resX;
                float y = (max2-pts[i][param2])/max2*(float)resY;
                if(mode == DOTPLOT) {
                    g2.draw(new Line2D.Float(x,y,x,y));
                } else if(mode == HIST) {
                    g2.draw(new Line2D.Float(lastX,lastY,x,y));
                    lastX = x; lastY = y;
                } else if(mode == SM_HIST) {
                    g2.draw(new Line2D.Float(lastX,lastY,x,y));
                    lastX = x; lastY = y;
                    i += 5;
                } else if(mode == DENSITY) {
                    int den = twoHist[pts[i][0]][pts[i][1]]; //density at cur pos
                    
                    if(den != 0) {
                        float ratio = Math.min(1f, (float)den/(float)dMax);
                        Color c = Color.getHSBColor(0.25f+0.125f+ratio,1f,1f);
                        g2.setColor(c);
                        g2.draw(new Line2D.Float(x,y,x,y));
                    }
                } else if(mode == CONTOUR) {
					g2.draw(new Line2D.Float(x,y,x,y));
                }
            }
        }
	}*/
    
    /*public int[][] genHistogram(int[][] data, int param) {
        int[][] h = new int[params[0][param].range][2];
        for(int i = 0; i < params[0][param].range; i++) {
            h[i][0] = i;
            h[i][1] = 0;
        }
        for(int i = 0; i < data.length; i++) {
            h[data[i][param]][1]++;
        }
        return h;
    }
    
    public int getMax(int[][] data) {
        int max = 0;
        for(int i = 0; i < data.length; i++) {
            if(data[i][1] > max) {
                max = data[i][1];
            }
        }
        if(hSecondMax) {
            int max2 = 0;
            for(int i = 0; i < data.length; i++) {
                if(data[i][1] > max2 && data[i][1] < max) {
                    max2 = data[i][1];
                }
            }
            return max2;
        }
        return max;
    }
    
    public int[][] ones(int param1, int param2, int smooth) {
        //okay, so it's not really ones, but it contains every point in the 2D plot
        int[][] a = new int[params[0][param1].range/smooth*params[0][param2].range/smooth][2];
        
        int n = 0;
        for(int i = 0; i < params[0][param1].range/smooth; i++) {
            for(int j = 0; j < params[0][param2].range/smooth; j++) {
                a[n][0] = i;
                a[n][1] = j;
                n++;
            }
        }
        return a;
    }
    
    public int[][] gen2DHist(int[][] data, int res, int smooth2) {
		int smooth = params[0][param1].range/res;
        int[][] h = new int[res][res];
        for(int i = 0; i < res; i++) {
            for(int j = 0; j < res; j++) {
                h[i][j] = 0;
            }
        }
        for(int i = 0; i < data.length; i++) {
            int val1 = (data[i][param1]/smooth)*1/smooth2;
			if(val1 < 0)
				val1 = 0;
            int val2 = (data[i][param2]/smooth)*1/smooth2;
			if(val2 < 0)
				val2 = 0;
            for(int j = 0; j < smooth2; j++) {
                for(int k = 0; k < smooth2; k++) {
                    h[val1*smooth2+j][val2*smooth2+k]++;
                }
            }
        }
        return h;
    }
    
    public int get2DMax(int[][] data) { //will always return 2nd highest
        int max = 0;
        int lastmax = 0;
        for(int i = 0; i < data.length; i++) {
            for(int j = 0; j < data[i].length; j++) {
                if(data[i][j] > max) {
                    lastmax = max;
                    max = data[i][j];
                }
            }
        }
        return lastmax;
    }*/

	public boolean atLevel(int[][] hist, int i, int j, int[] levels) {
		for(int l = 0; l < levels.length; l++) {
			int error = (int)((double)levels[l] * 0.3);
			//if(i >= levels[l]-error && i <= levels[l]+error)
			//	return true;
			if(isIsect(hist, i, j, levels[l]))
				return true;
		}

		return false;
	}
    
    public int[][] contour(int[][] hist, int[] levels, boolean outliers) {
		outliers = false;

		Vector c = new Vector();

		for(int i = 0; i < hist.length; i++) {
			for(int j = 0; j < hist[i].length; j++) {
				if(hist[i][j] > 0 && hist[i][j] < levels[0] && outliers) {//outliers
					int[] p = { i, j };
					c.add(p);
				} else if(atLevel(hist,i,j,levels)) {
					int[] p = { i, j };
					c.add(p);
				}
			}
		}
        
        int[][] a = new int[c.size()][2];
        for(int i = 0; i < c.size(); i++) {
            int[] p = (int[])c.elementAt(i);
            a[i][0] = p[0];
            a[i][1] = p[1];
        }

        return a;
    }
    
    public boolean isIsect(int[][] hist, int i, int j, int level) {
        if(i != 0 && ((hist[i][j] > level && hist[i-1][j] < level)
            || (hist[i][j] < level && hist[i-1][j] > level))) {
                return true;
        }
        if(i != hist.length-1 && ((hist[i][j] > level && hist[i+1][j] < level)
            || (hist[i][j] < level && hist[i+1][j] > level))) {
                return true;
        }
        if(j != 0 && i != 0 && ((hist[i][j] > level && hist[i-1][j-1] < level)
            || (hist[i][j] < level && hist[i-1][j-1] > level))) {
                return true;
        }
        if(j != hist[0].length-1 && i != 0 && ((hist[i][j] > level && hist[i-1][j+1] < level)
            || (hist[i][j] < level && hist[i-1][j+1] > level))) {
                return true;
        }
        if(i != hist.length-1 && j != 0 && ((hist[i][j] > level && hist[i+1][j-1] < level)
            || (hist[i][j] < level && hist[i+1][j-1] > level))) {
                return true;
        }
        if(j != hist[0].length-1 && i != hist.length-1 && ((hist[i][j] > level && hist[i+1][j+1] < level)
            || (hist[i][j] < level && hist[i+1][j+1] > level))) {
                return true;
        }
        if(j != 0 && ((hist[i][j] > level && hist[i][j-1] < level)
            || (hist[i][j] < level && hist[i][j-1] > level))) {
                return true;
        }
        if(j != hist[0].length-1 && ((hist[i][j] > level && hist[i][j+1] < level)
            || (hist[i][j] < level && hist[i][j+1] > level))) {
                return true;
        }
        return false;
    }
    
    public boolean inList(Vector v, int[] p) {
        for(int i = 0; i < v.size(); i++) {
            int[] vi = (int[])v.elementAt(i);
            if(vi[0] == p[0] && vi[1] == p[1])
                return true;
        }
        return false;
    }

    public void paintComponent(Graphics g) {
        Graphics2D g2 = (Graphics2D)g;
        
        //draw background and frame
        g2.setColor(Color.white);
        g2.fillRect(0,0,getWidth()-1,getHeight()-1);
        
        g2.setColor(Color.black);
        g2.drawRect(0,0,getWidth()-1,getHeight()-1);
        if(selected)
            g2.drawRect(1,1,getWidth()-3,getHeight()-3);
        
        if(!draw)
            return;
        
        g2.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
        //g2.setRenderingHint(RenderingHints.KEY_RENDERING, RenderingHints.VALUE_RENDER_QUALITY);
        
        //draw actual plot
        w = (double)(getWidth()-axisBorder-2*frameBorder);
        h = (double)(getHeight()-axisBorder-2*frameBorder-title);
        AffineTransform at = AffineTransform.getScaleInstance(w/(double)resX,h/(double)resY);
		
		g2.drawImage(bi,new AffineTransformOp(at,g2.getRenderingHints()),frameBorder+axisBorder,frameBorder+title);
        
        g2.setColor(Color.black);
        //g2.drawRect(axisBorder+frameBorder-1, frameBorder+title, (int)w, (int)h);
		g2.draw(new Rectangle2D.Double(axisBorder+frameBorder, frameBorder+title, w, h));
        
        //draw title
        Font f = new Font("Serif", Font.BOLD, 13);
        g2.setFont(f);
        FontMetrics fm = g2.getFontMetrics();
       
        String old = "", stitle = "";
        boolean okay = false;
        for(int i = 0; i < fData.length; i++) {
            okay = false;
            stitle += fData[i].name;
            if(i != fData.length-1)
                stitle+=", ";
            if(fm.stringWidth(stitle+", ...") > getWidth())
                break;
            okay = true;
            old = stitle;
        }
        if(!okay && fData.length != 1)
            old += "...";
        g2.drawString(old, getWidth()/2-fm.stringWidth(old)/2, fm.getHeight());
        
        //draw axis labels
        float n = 20;
        f = new Font("Serif", Font.PLAIN, (int)n);
        fm = g2.getFontMetrics(f.deriveFont(n));
        while(fm.getHeight() > axisBorder) {
            n--;
            fm = g2.getFontMetrics(f.deriveFont(n));
        }

		float nLabel = 12;
        fm = g2.getFontMetrics(f.deriveFont(nLabel));
        while(fm.getMaxAscent() > frameBorder-tickSize) {
            nLabel--;
            fm = g2.getFontMetrics(f.deriveFont(nLabel));
        }

        g2.setFont(f.deriveFont(n));
        fm = g2.getFontMetrics();
        String axis = params[0][param1].lname; //x-axis
        g2.drawString(axis,axisBorder+(int)(w/2)-fm.stringWidth(axis)/2,axisBorder+(int)h+2*frameBorder+title-fm.getMaxDescent());
        
        if(param2 != -1)
            axis = params[0][param2].lname; //y-axis
        else
            axis = "Percent Max";
        g2.rotate(3*Math.PI/2);
		g2.drawString(axis,-getHeight()+(axisBorder-frameBorder+title+(int)(h/2)-fm.stringWidth(axis)/2),fm.getMaxAscent());
		if(!vectorOutput) {
	        g2.rotate(-3*Math.PI/2);
		} else {
			g2.rotate(0); //key to "</g>" in SVG
		}

        /**draw tick marks**/
		g2.setFont(f.deriveFont(nLabel));
        fm = g2.getFontMetrics();
        //x-axis
        if(log1) {
            int dec = params[0][param1].amp1;
            double scale = w / (double)dec;
            scale /= Math.log(10);
            
            for(int i = 1; i <= dec; i++) {
                double prod = Math.pow(10., (double)i);
                double delta = prod/10;
                for(int j = 1; j <= 10; j++) {
                    g2.draw(new Line2D.Double(axisBorder+frameBorder+Math.log(j*delta)*scale,h+frameBorder+title,axisBorder+frameBorder+Math.log(j*delta)*scale,h+title+frameBorder+tickSize));
					if(j==1) {//label for first in each decade
						String l;
						if(i==1)
							l="1";
						else
							l="1E"+(i-1);
						g2.drawString(l, axisBorder+frameBorder+(int)(Math.log(j*delta)*scale), (int)h+frameBorder+title+tickSize+fm.getMaxAscent());
					}
					if(i==dec && j==10) {//label for last decade (max pt)
						String l = "1E"+(i);
						g2.drawString(l, axisBorder+frameBorder+(int)(Math.log(j*delta)*scale)-fm.stringWidth(l)+2, (int)h+frameBorder+title+tickSize+fm.getMaxAscent());
					}
                }
            }
            
        } else { //Linear
            double delta = w/xTicks;
            for(int i = 0; i <= xTicks; i++) {
                g2.draw(new Line2D.Double(axisBorder+frameBorder+i*delta,h+frameBorder+title,axisBorder+frameBorder+i*delta,h+title+frameBorder+tickSize));
				if(i == 0) //numerical label at beginning
					g2.drawString("0", axisBorder+frameBorder+(int)(i*delta), (int)h+frameBorder+title+tickSize+fm.getMaxAscent());
				if(i == xTicks) {//numerical label at end
					String l = new Integer(params[0][param1].range).toString();
					g2.drawString(l, axisBorder+frameBorder+(int)(i*delta)-fm.stringWidth(l)+2, (int)h+frameBorder+title+tickSize+fm.getMaxAscent());
				}
            }
        }
        //y-axis
        if(log2) {
            int dec = params[0][param2].amp1;
            double scale = h / (double)dec;
            scale /= Math.log(10);
            
            for(int i = 1; i <= dec; i++) {
                double prod = Math.pow(10., (double)i);
                double delta = prod/10;
                for(int j = 1; j <= 10; j++) {
                    g2.draw(new Line2D.Double(axisBorder+(frameBorder-tickSize),h+frameBorder+title-Math.log(j*delta)*scale,axisBorder+frameBorder,h+frameBorder+title-Math.log(j*delta)*scale));
					if(j==1 || (i==dec&&j==10)) {//label for first in each decade, and last one
						String l;
						if(i==1)
							l="1";
						else if(j!=10)
							l="1E"+(i-1);
						else
							l="1E"+i;
						g2.drawString(l, axisBorder+frameBorder-fm.stringWidth(l), (int)h+frameBorder+title-(int)(Math.log(j*delta)*scale)-fm.getDescent());
					}
                }
            }
            
        } else { //Linear
            double delta = h/yTicks;
            for(int i = 0; i <= yTicks; i++) {
                g2.draw(new Line2D.Double(axisBorder+(frameBorder-tickSize),h+frameBorder+title-i*delta,axisBorder+frameBorder,h+frameBorder+title-i*delta));
				if(i == 0 || i == yTicks) { //numerical label
					String l;
					if(i==0) {
						l="0";
					} else {
						if(param2 != -1) {
							l = new Integer(params[0][param2].range).toString();
						} else { //is a histogram
							l = "100";
						}
					}
					g2.drawString(l, axisBorder+frameBorder-fm.stringWidth(l), (int)h+frameBorder+title-(int)(i*delta)-fm.getDescent());
				}
            }
        }
        
        //regions
        Stroke s = g2.getStroke();
        for(int i = 0; i < regions.size(); i++) {
            double[] r = ((Region)regions.elementAt(i)).getVals(fData[0].filename);
            if(mode != DENSITY)
            	g2.setColor(Color.green);
            else
            	g2.setColor(Color.black);
            if(r[1] != param1 || r[2] != param2)
                continue;
            if(r[0] < 0)
                continue;
            if(regionEditing == i)
                g2.setStroke(new BasicStroke(2));
            else
                g2.setStroke(s);
            if((int)r[0] == ELLIPSE) {
                Ellipse2D.Double el = new Ellipse2D.Double(0,-r[6]/2,r[5],r[6]);
                AffineTransform atEl = AffineTransform.getTranslateInstance(r[3],r[4]+r[6]/2);
                atEl.rotate(r[7]);
                Shape sEl = atEl.createTransformedShape(el);
                g2.draw(sEl);
                g2.setColor(new Color(0,0,0,0.1f));
                g2.fill(sEl);
                if(regionEditing == i) { //draw square things
                    g2.setColor(new Color(0,0,0,0.5f));
                    Rectangle2D.Double sq = new Rectangle2D.Double(-gateDragBoxSize,-gateDragBoxSize,2*gateDragBoxSize,2*gateDragBoxSize);
                    Shape trans = atEl.createTransformedShape(sq);
                    g2.fill(trans);
                    sq = new Rectangle2D.Double(r[5]-gateDragBoxSize,-gateDragBoxSize,2*gateDragBoxSize,2*gateDragBoxSize);
                    trans = atEl.createTransformedShape(sq);
                    g2.fill(trans);
                    sq = new Rectangle2D.Double(r[5]/2-gateDragBoxSize,-r[6]/2-gateDragBoxSize,2*gateDragBoxSize,2*gateDragBoxSize);
                    trans = atEl.createTransformedShape(sq);
                    g2.fill(trans);
                    sq = new Rectangle2D.Double(r[5]/2-gateDragBoxSize,r[6]/2-gateDragBoxSize,2*gateDragBoxSize,2*gateDragBoxSize);
                    trans = atEl.createTransformedShape(sq);
                    g2.fill(trans);
                }
            } else if((int)r[0] == RECT) {
                Rectangle2D.Double rect = new Rectangle2D.Double(r[3],r[4],r[5],r[6]);
                g2.draw(rect);
                g2.setColor(new Color(0,0,0,0.1f));
                g2.fill(rect);
                if(regionEditing == i) { //draw square things
                    g2.setColor(new Color(0,0,0,0.5f));
                    Rectangle2D.Double sq = new Rectangle2D.Double(r[3]-gateDragBoxSize,r[4]-gateDragBoxSize,2*gateDragBoxSize,2*gateDragBoxSize);
                    g2.fill(sq);
                    sq = new Rectangle2D.Double(r[3]-gateDragBoxSize,r[4]+r[6]-gateDragBoxSize,2*gateDragBoxSize,2*gateDragBoxSize);
                    g2.fill(sq);
                    sq = new Rectangle2D.Double(r[3]+r[5]-gateDragBoxSize,r[4]-gateDragBoxSize,2*gateDragBoxSize,2*gateDragBoxSize);
                    g2.fill(sq);
                    sq = new Rectangle2D.Double(r[3]+r[5]-gateDragBoxSize,r[4]+r[6]-gateDragBoxSize,2*gateDragBoxSize,2*gateDragBoxSize);
                    g2.fill(sq);
                }
            } else if((int)r[0] == QUAD) {
                Line2D.Double line = new Line2D.Double(r[3],frameBorder+title,r[3],frameBorder+h+title);
                g2.draw(line);
                line = new Line2D.Double(axisBorder+frameBorder,r[4],axisBorder+frameBorder+w,r[4]);
                g2.draw(line);
                if(regionEditing == i) { //draw square things
                    g2.setColor(new Color(0,0,0,0.5f));
                    Rectangle2D.Double sq = new Rectangle2D.Double(r[3]-gateDragBoxSize,r[4]-gateDragBoxSize,2*gateDragBoxSize,2*gateDragBoxSize);
                    g2.fill(sq);
                }
            } else if((int)r[0] == POLY) {
                double x, y;
                double xold = 0;
                double yold = 0;
                for(int p = 3; p < r.length; p++) {
                    x = r[p];
                    p++;
                    y = r[p];
                    if(xold == 0 && yold == 0) {
                        g2.draw(new Line2D.Double(x,y,x,y));
                    } else {
                        if(x != 0 && y != 0)
                            g2.draw(new Line2D.Double(xold,yold,x,y));
                    }
                    xold = x; yold = y;
                    
                    Color old_color = g2.getColor();
                    if(regionEditing == i) { //draw square things
                        g2.setColor(new Color(0,0,0,0.5f));
                        Rectangle2D.Double sq = new Rectangle2D.Double(x-gateDragBoxSize,y-gateDragBoxSize,2*gateDragBoxSize,2*gateDragBoxSize);
                        g2.fill(sq);
                    }
                    g2.setColor(old_color);
                }
                
                if(regionEditing != i || region == NONE)
                    g2.draw(new Line2D.Double(r[r.length-2],r[r.length-1],r[3],r[4]));
                
                int[] xp = new int[(r.length-3)/2];
                int[] yp = new int[(r.length-3)/2];
                int np = 0;
                for(int p = 3; p < r.length; p++) {
                    xp[np] = (int)r[p];
                    p++;
                    yp[np] = (int)r[p];
                    np++;
                }
                if(xp[np-1] == 0 && yp[np-1] == 0)
                    np--;
                
                Polygon poly = new Polygon(xp,yp,np);
                g2.setColor(new Color(0,0,0,0.1f));
                g2.fill(poly);
            } else if((int)r[0] == RANGE) {
                double theta = Math.atan((r[6]-r[4])/(r[5]-r[3]));
                if((theta < Math.PI/5 && theta > -Math.PI/5) || mode == HIST || mode == SM_HIST) {
                    g2.draw(new Line2D.Double(r[3],r[4],r[5],r[4])); //horiz
                    g2.draw(new Line2D.Double(r[3],r[4]-frameBorder-2,r[3],r[4]+frameBorder+2));
                    g2.draw(new Line2D.Double(r[5],r[4]-frameBorder-2,r[5],r[4]+frameBorder+2));
                    if(regionEditing == i) { //draw square things
                        g2.setColor(new Color(0,0,0,0.5f));
                        Rectangle2D.Double sq = new Rectangle2D.Double(r[3]-gateDragBoxSize,r[4]-gateDragBoxSize,2*gateDragBoxSize,2*gateDragBoxSize);
                        g2.fill(sq);
                        sq = new Rectangle2D.Double(r[5]-gateDragBoxSize,r[4]-gateDragBoxSize,2*gateDragBoxSize,2*gateDragBoxSize);
                        g2.fill(sq);
                    }
                } else {
                    g2.draw(new Line2D.Double(r[3],r[4],r[3],r[6])); //vert
                    g2.draw(new Line2D.Double(r[3]-frameBorder-2,r[4],r[3]+frameBorder+2,r[4]));
                    g2.draw(new Line2D.Double(r[3]-frameBorder-2,r[6],r[3]+frameBorder+2,r[6]));
                    if(regionEditing == i) { //draw square things
                        g2.setColor(new Color(0,0,0,0.5f));
                        Rectangle2D.Double sq = new Rectangle2D.Double(r[3]-gateDragBoxSize,r[4]-gateDragBoxSize,2*gateDragBoxSize,2*gateDragBoxSize);
                        g2.fill(sq);
                        sq = new Rectangle2D.Double(r[3]-gateDragBoxSize,r[6]-gateDragBoxSize,2*gateDragBoxSize,2*gateDragBoxSize);
                        g2.fill(sq);
                    }
                }
            }
            //percent labels
            /*if(fData.length == 1) {
                if(i < percents.size()) { //not still editing
                    double[] label = (double[])percents.elementAt(i);
                    if(label[0] == 1) {
                        for(int j = 1; j < label.length; j++) {
                            g2.setColor(new Color(1,1,1,0.7f));
                            String l = (int)(label[j]*100.)+"%";
                            g2.setFont(new Font("SansSerif", Font.PLAIN, 13));
                            j++;
                            double newX = label[j]-g2.getFontMetrics().stringWidth(l)/2;
                            j++;
                            double newY = label[j];
                            Rectangle2D.Double sq = new Rectangle2D.Double(newX-2,newY-g2.getFontMetrics().getHeight()/1.5-1,g2.getFontMetrics().stringWidth(l)+4,g2.getFontMetrics().getAscent()+2);
                            g2.fill(sq);
                            g2.setColor(Color.black);
                            
                            g2.drawString(l,(int)newX,(int)newY);
                        }
                    }
                }
            }*/
        }
        g2.setStroke(s);
    }
    
    
    public Dimension getMinimumSize() {
        return getPreferredSize();
    }
    
    public Dimension getMaximumSize() {
        return getPreferredSize();
    }
    
    public Dimension getPreferredSize() {
        return preferredSize;
    }
    
    public static Dimension getPSize() {
    	return preferredSize;
    }
    
    public boolean mouseIn(MouseEvent e) {
        w = (double)(getWidth()-axisBorder-2*frameBorder);
        h = (double)(getHeight()-axisBorder-2*frameBorder-title);
        
        if(region != NONE)
            return true;
        if(regionHere(e) != -1)
            return true;
        if(e.isAltDown() || e.getButton() == MouseEvent.BUTTON3)
            return true;
        if(e.getX() < axisBorder && e.getY() < frameBorder+h) //y-axis
            return true;
        else if(e.getY() > 2*frameBorder+h+title && e.getX() > axisBorder+frameBorder) //x-axis
            return true;
        return false;
    }
    
    public int regionHere(MouseEvent e) {
        for(int i = 0; i < regions.size(); i++) {
            double[] r = ((Region)regions.elementAt(i)).getVals(fData[0].filename);
            if(r[1] != param1 || r[2] != param2)
                continue;
            if((int)r[0] == RECT) {
                Rectangle2D.Double rect = new Rectangle2D.Double(r[3],r[4],r[5],r[6]);
                if(rect.contains((double)e.getX(),(double)e.getY()))
                    return i;
            } else if((int)r[0] == ELLIPSE) {
                Ellipse2D.Double el = new Ellipse2D.Double(0,-r[6]/2,r[5],r[6]);
                AffineTransform atEl = AffineTransform.getTranslateInstance(r[3],r[4]+r[6]/2);
                atEl.rotate(r[7]);
                Shape sEl = atEl.createTransformedShape(el);
                if(sEl.contains((double)e.getX(),(double)e.getY()))
                    return i;
            } else if((int)r[0] == RANGE) {
                
                double theta = Math.atan((r[6]-r[4])/(r[5]-r[3]));
                Rectangle2D.Double rect;
                if(mode == HIST || mode == SM_HIST || (theta < Math.PI/5 && theta > -Math.PI/5)) {
                    //horiz
                    rect = new Rectangle2D.Double(r[3],r[4]-frameBorder,r[5]-r[3],2*frameBorder);
                } else {
                    //vert
                    rect = new Rectangle2D.Double(r[3]-frameBorder,r[4],2*frameBorder,r[6]-r[4]);
                }
                if(rect.contains((double)e.getX(),(double)e.getY()))
                    return i;
            } else if((int)r[0] == QUAD) {
                Rectangle2D.Double rect = new Rectangle2D.Double(r[3]-frameBorder,r[4]-frameBorder,2*frameBorder,2*frameBorder);
                if(rect.contains((double)e.getX(),(double)e.getY()))
                    return i;
            } else if((int)r[0] == POLY) {
                int[] xp = new int[(r.length-3)/2];
                int[] yp = new int[(r.length-3)/2];
                int n = 0;
                for(int p = 3; p < r.length; p++) {
                    xp[n] = (int)r[p];
                    p++;
                    yp[n] = (int)r[p];
                    n++;
                }
                Polygon poly = new Polygon(xp,yp,(r.length-3)/2);
                if(poly.contains((double)e.getX(),(double)e.getY()))
                    return i;
            }
        }
        
        return -1;
    }
    
    public void mouseClicked(MouseEvent e) {
        if(!mouseIn(e)) { //it's an event for the PlotPanel
            regionEditing = -1;
            rStartX = -1;
            rStartY = -1;
            region = NONE;
            repaint();
            /*e.translatePoint((int)getLocation().getX(),(int)getLocation().getY());
            pp.mouseClicked(e);*/
            return;
        }
                
        if(region == NONE) {
            int rEdit = regionHere(e);
            if(rEdit != -1) {
				if(!((Boolean)rLocked.elementAt(rEdit)).booleanValue()) {
					regionEditing = rEdit;
					grabFocus();
					repaint();
				}
            }
            
            if(e.getClickCount() == 2) {
                double type = ((Region)regions.elementAt(rEdit)).getVals(fData[0].filename)[0];
                if(!((int)type == QUAD) && !((Boolean)rLocked.elementAt(rEdit)).booleanValue()) {
                    /*Plot newP = new Plot(pp,facs);
                    ToolTipManager.sharedInstance().registerComponent(newP);
                    newP.gates.addAll(gates); //inherit from parent
                    newP.gates.add(new Gate(dataIndices,regionEditing,(String)facs.rNames.elementAt(regionEditing)));
                    //newP.rDisplayed = (Vector)rDisplayed.clone();
                    //newP.rDisplayed.setElementAt(new Boolean(false), regionEditing);
                    newP.setPlot(dataIndices,facs.gate(fData,newP.gates), param1, param2);
                    
                    Insets insets = pp.getInsets();
                    Dimension size = newP.getPreferredSize();
                    newP.setBounds(5 + insets.left, 5 + insets.top, size.width, size.height);
                    
                    pp.add(newP,0);
                    pp.validate(); pp.repaint();*/
					gates.add(new Gate(facs.getFDataNames(dataIndices),regionEditing,((Region)facs.regions.elementAt(regionEditing)).getName()));
					rDisplayed.setElementAt(new Boolean(false),regionEditing);
					setPlot(dataIndices,facs.gate(dataIndices,fData,gates), param1, param2, mode);
                }
            }
        }
        
        if(region == POLY) {
            if(e.getClickCount() == 2) {
                if(regionEditing != -1) {
                    //double click registers first as a single click...
                    double[] r = ((Region)regions.elementAt(regionEditing)).getVals(fData[0].filename);
                    double[] r2 = new double[r.length-2];
                    for(int i = 0; i < r.length-2; i++) {
                        r2[i] = r[i];
                    }
                    ((Region)regions.elementAt(regionEditing)).setVals(fData[0].filename,r2);
                    
                    if(defaultRegionName == null) {
	                    String name = JOptionPane.showInputDialog(null,"Name New Gate",((Region)regions.elementAt(regionEditing)).getName());
	                	((Region)regions.elementAt(regionEditing)).setName(name);
	                }
                    
                    Region unT = unTranslate((Region)regions.elementAt(regionEditing));
                    facs.regions.add(unT);
                    
                    //percents.add(generatePctLabel(regions,regionEditing));
                    
                    facs.updatePlots(regionEditing,true);
                }
                
                rStartX = -1;
                rStartY = -1;
                regionEditing = -1;
                region = NONE;
                repaint();
                setCursor(null);
            } else {
                rStartX = e.getX();
                rStartY = e.getY();
                double[] r = { (double)POLY, param1, param2, 0.0, 0.0 }; //x, y, x, y
                if(regionEditing == -1) {
                    regionEditing = regions.size();
                    regions.add(new Region(r,makeRName("G"+regionEditing))); //just so we can do setElementAt
                    rDisplayed.add(new Boolean(true));
					rLocked.add(new Boolean(false));
                } else {
                    r = ((Region)regions.elementAt(regionEditing)).getVals(fData[0].filename);
                }
                r[r.length-2] = e.getX();
                r[r.length-1] = e.getY();
                double[] r2 = new double[r.length+2];
                for(int i = 0; i < r.length; i++) {
                    r2[i] = r[i];
                }
                r2[r.length] = 0.0;
                r2[r.length+1] = 0.0;
				((Region)regions.elementAt(regionEditing)).setVals(fData[0].filename,r2);
                repaint();
            }
        }
        
        if(e.getX() < axisBorder && e.getY() < frameBorder+h+title) { //y-axis
            if(!draw)
                return;
            
            JPopupMenu popChan = new JPopupMenu();
            String t = "Histogram";
            if(param2 == -1)
                t = ">"+t;
            JMenuItem hist = new JMenuItem(t);
            hist.addActionListener(this);
            hist.setActionCommand("Y:"+(-1));
            popChan.add(hist);
            
            JMenuItem[] mis = new JMenuItem[params[0].length];
            for(int i = 0; i < params[0].length; i++) {
                t = params[0][i].lname;
                if(i == param2)
                    t = ">"+t;
                mis[i] = new JMenuItem(t);
                mis[i].addActionListener(this);
                mis[i].setActionCommand("Y:"+i);
                popChan.add(mis[i]);
            }
            popChan.show(this, e.getX(), e.getY());
        } else if(e.getY() > 2*frameBorder+h+title && e.getX() > axisBorder+frameBorder) { //x-axis
            if(!draw)
                return;
            
            JPopupMenu popChan = new JPopupMenu();
            JMenuItem[] mis = new JMenuItem[params[0].length];
            for(int i = 0; i < params[0].length; i++) {
                String t = params[0][i].lname;
                if(i == param1)
                    t = ">"+t;
                mis[i] = new JMenuItem(t);
                mis[i].addActionListener(this);
                mis[i].setActionCommand("X:"+i);
                popChan.add(mis[i]);
            }
            popChan.show(this, e.getX(), e.getY());
        //} else if(e.getX() > frameBorder && e.getX() < frameBorder+h && e.getY() > axisBorder+frameBorder+title && e.getY() < axisBorder+frameBorder+w) {
        } else {
            //click inside
            if(e.isAltDown() || e.getButton() == MouseEvent.BUTTON3) {
                //right-click or Alt-click
                JPopupMenu popOpt = new JPopupMenu();
                JMenuItem miData = new JMenuItem("Choose Data...");
                miData.setActionCommand("data");
                miData.addActionListener(this);
                popOpt.add(miData);
                JMenuItem miGates = new JMenuItem("Show/Hide Gates...");
                miGates.setActionCommand("gates");
                miGates.addActionListener(this);
                if(!draw)
                    miGates.setEnabled(false);
                popOpt.add(miGates);
                popOpt.addSeparator();
                
                if(draw) {
                    if(param2 == -1) { //histogram
                        popOpt.add(cbSmooth);
                        popOpt.add(cb2Max);
                    } else {
                        popOpt.add(rbDot);
                        popOpt.add(rbDensity);
                        popOpt.add(rbContour);
                        if(fData.length != 1) {
                            rbDensity.setEnabled(false);
                            rbContour.setEnabled(false);
                        } else {
                            rbDensity.setEnabled(true);
                            rbContour.setEnabled(true);
                        }
                        popOpt.add(new JSeparator());
                        if(mode == CONTOUR)
                        	popOpt.add(cbOutliers);
                        if(mode == DENSITY)
                        	popOpt.add(cbDSmooth);
                    }
                }
                popOpt.show(this, e.getX(), e.getY());
            }
        }
        repaint();
    }

	public String makeRName(String in) { //ensure all region names are unique
		if(defaultRegionName != null)
			in = defaultRegionName;
		for(int i = 0; i < regions.size(); i++) {
			if(((Region)regions.elementAt(i)).getName().equals(in)) {
				in += "'";
				i = 0;
			}
		}
		return in;
	}
    
    public void actionPerformed(ActionEvent e) {
        if(e.getActionCommand().startsWith("X:")) {
            int p = new Integer(e.getActionCommand().substring(2)).intValue();
            setParams(p,param2,mode);
        } else if(e.getActionCommand().startsWith("Y:")) {
            int p = new Integer(e.getActionCommand().substring(2)).intValue();
            if(p == -1)
                mode = old1D;
            else
                mode = old2D;
            setParams(param1,p,mode);
        } else if(e.getActionCommand().equals("density")) {
            cbOutliers.setEnabled(false);
            setParams(param1,param2,DENSITY);
            old2D = mode;
        } else if(e.getActionCommand().equals("dot")) {
            cbOutliers.setEnabled(false);
            setParams(param1,param2,DOTPLOT);
            old2D = mode;
        } else if(e.getActionCommand().equals("contour")) {
            cbOutliers.setEnabled(true);
            setParams(param1,param2,CONTOUR);
            old2D = mode;
        } else if(e.getActionCommand().equals("data")) {
            int[] f;
            FACSData[] sel;
            if(fData == null) {
                sel = new FACSData[0];
            } else {
                sel = fData;
            }
            
            Vector ddgates = new Vector();
            for(int i = 0; i < regions.size(); i++) {
                String[] datas = new String[0];
                for(int j = 0; j < gates.size(); j++) {
                    if(((Gate)gates.elementAt(j)).region == i) {
                        datas = ((Gate)gates.elementAt(j)).data;
                        break;
                    }
                }
                
                ddgates.add(new Gate(datas, i, ((Region)facs.regions.elementAt(i)).getName()));
            }
            
            DataDialog dd = new DataDialog(this, sel, facs.data, ddgates);
            
            if(dd.okay && dd.added.size() > 0) {
                f = DataDialog.makeArray(dd.added);
				for(int i = 0; i < dd.gates.size(); i++) {
					if(((Gate)dd.gates.elementAt(i)).data.length > 0)
						rDisplayed.setElementAt(new Boolean(false),((Gate)dd.gates.elementAt(i)).region);
				}
                setPlot(f, facs.gate(f,facs.getFDatas(f),dd.gates), param1, param2, mode);
                gates = dd.gates;
                
                //legend for plots with mult datas
                /*int found = -1;
                for(int i = 0; i < pp.getKeys().size(); i++) {
                    PlotKey pk = (PlotKey)pp.getKeys().elementAt(i);
                    if(pk.plot == this) {
                        found = i;
                        break;
                    }
                }*/
				boolean found = false;
				if(facs.pk != null)
					found = true;
                if(dd.added.size() > 1 || found) {
                    
                    if(!found) {
                        Hashtable datas = new Hashtable();
                        Vector dnames = new Vector();
                        for(int i = 0; i < fData.length; i++) {
                            datas.put(fData[i].name, colors[i]);
                            dnames.add(fData[i].name);
                        }
                        facs.pk = new PlotKey(dnames,datas,this);
                        
                        //Insets insets = pp.getInsets();
                        //Dimension size = facs.pk.getPreferredSize();
                        //pk.setBounds(5 + insets.left + getX() + getWidth(), insets.top + getY(), size.width, size.height);
                        
                        //pp.add(pk,0);
                        facs.pk.repaint();
                    } else {
                        Hashtable datas = new Hashtable();
                        Vector dnames = new Vector();
                        for(int i = 0; i < fData.length; i++) {
                            datas.put(fData[i].name, colors[i]);
                            dnames.add(fData[i].name);
                        }
                        facs.pk.datas = datas;
                        facs.pk.dnames = dnames;
                        facs.pk.repaint();
                    }
                }
            }
        } else if(e.getActionCommand().equals("gates")) {
            RegionsDialog rd = new RegionsDialog(true,fData[0].filename,regions,rDisplayed,rLocked,fData[0].params);
            if(rd.okay) {
                rDisplayed = rd.rDisplayed;
				rLocked = rd.rLocked;
            }
			setVisibilities();
			facs.regions = unTranslateAll(regions);
			facs.updatePlots(0,true);
        }
    }
    
    public void setVisibilities() {
        for(int i = 0; i < regions.size(); i++) {
            double[] r = ((Region)regions.elementAt(i)).getVals(fData[0].filename);
            r[0] = Math.abs(r[0]);
            if(!((Boolean)rDisplayed.elementAt(i)).booleanValue())
                r[0] = -r[0];
			((Region)regions.elementAt(i)).setVals(fData[0].filename,r);
        }
        repaint();
    }
    
    public void unlockAll() {
    	for(int i = 0; i < rLocked.size(); i++) {
    		rLocked.setElementAt(new Boolean(false),i);
    	}
    }
    
    public void mouseEntered(MouseEvent e) {
        if(!mouseIn(e)) { //it's an event for the PlotPanel
            /*e.translatePoint((int)getLocation().getX(),(int)getLocation().getY());
            pp.mouseEntered(e);*/
            return;
        }
    }
    
    public void mouseExited(MouseEvent e) {
        if(!mouseIn(e)) { //it's an event for the PlotPanel
            /*e.translatePoint((int)getLocation().getX(),(int)getLocation().getY());
            pp.mouseExited(e);*/
            return;
        }
        if(region == QUAD) {
            if(regionEditing != -1) {
                regions.removeElementAt(regionEditing);
                regionEditing = -1;
            }
            repaint();
        }
    }
    
    public void mousePressed(MouseEvent e) {
        if(!mouseIn(e) && regionEditing == -1) { //it's an event for the PlotPanel
            /*e.translatePoint((int)getLocation().getX(),(int)getLocation().getY());
            pp.mousePressed(e);*/
            return;
        }
        if(!draw)
            return;
        
        //region editing / creation
        if(region != NONE) {
            if(e.getX() > axisBorder+frameBorder && e.getX() <= axisBorder+frameBorder+w &&
                    e.getY() > frameBorder+title && e.getY() < frameBorder+title+h) {
                if(mode != HIST && mode != SM_HIST && region != POLY) {
                    //histograms only do range
                    //polygon handled through click and drag
                    rStartX = e.getX();
                    rStartY = e.getY();
                    
                
                    if(region == ELLIPSE) {
                        regionEditing = regions.size();
                        double[] el = { (double)Plot.ELLIPSE, param1, param2, (double)rStartX, (double)rStartY, 0.0, 0.0, 0.0 }; //x, y, w, h, theta
                        regions.add(new Region(el,makeRName("G"+regionEditing)));
                        rDisplayed.add(new Boolean(true));
						rLocked.add(new Boolean(false));
                    } else if(region == RECT) {
                        regionEditing = regions.size();
                        double[] r = { (double)Plot.RECT, param1, param2, (double)rStartX, (double)rStartY, 0.0, 0.0 }; //x, y, w, h
                        regions.add(new Region(r,makeRName("G"+regionEditing)));
                        rDisplayed.add(new Boolean(true));
						rLocked.add(new Boolean(false));
                    } else if(region == RANGE) {
                        regionEditing = regions.size();
                        double[] r = { (double)Plot.RANGE, param1, param2, (double)rStartX, (double)rStartY, (double)rStartX, (double)rStartY }; //x, y, x, y
                        regions.add(new Region(r,makeRName("G"+regionEditing)));
                        rDisplayed.add(new Boolean(true));
						rLocked.add(new Boolean(false));
                    }
                } else if(region == RANGE) {
                    rStartX = e.getX();
                    rStartY = e.getY();
                    regionEditing = regions.size();
                    double[] r = { (double)Plot.RANGE, param1, param2, (double)rStartX, (double)rStartY, (double)rStartX, (double)rStartY }; //x, y, x, y
                    regions.add(new Region(r,makeRName("G"+regionEditing)));
                    rDisplayed.add(new Boolean(true));
					rLocked.add(new Boolean(false));
                }
                repaint();
            }
        } else if(regionEditing != -1) { //editing, not creating
            int p = getParamEdit(e,regionEditing);
            if(p == -1) {
                regionEditing = -1;
                repaint();
            } else {
                paramEdit = p;
                rStartX = e.getX();
                rStartY = e.getY();
            }
        }
    }
    
    public int getParamEdit(MouseEvent e, int region) {
        double[] r = ((Region)regions.elementAt(region)).getVals(fData[0].filename);
        if((int)r[0] == RECT) {
            Rectangle2D.Double cur;
            cur = new Rectangle2D.Double(r[3]-frameBorder,r[4]-frameBorder,2*frameBorder,2*frameBorder);
            if(cur.contains((double)e.getX(),(double)e.getY()))
                return 1;
            
            cur = new Rectangle2D.Double(r[3]-frameBorder,r[4]+r[6]-frameBorder,2*frameBorder,2*frameBorder);
            if(cur.contains((double)e.getX(),(double)e.getY()))
                return 2;
            
            cur = new Rectangle2D.Double(r[3]+r[5]-frameBorder,r[4]-frameBorder,2*frameBorder,2*frameBorder);
            if(cur.contains((double)e.getX(),(double)e.getY()))
                return 3;
            
            cur = new Rectangle2D.Double(r[3]+r[5]-frameBorder,r[4]+r[6]-frameBorder,2*frameBorder,2*frameBorder);
            if(cur.contains((double)e.getX(),(double)e.getY()))
                return 4;
            
            cur = new Rectangle2D.Double(r[3],r[4],r[5],r[6]);
            if(cur.contains((double)e.getX(),(double)e.getY()))
                return 0;
            
            return -1;
        } else if((int)r[0] == ELLIPSE) {
            Shape cur;
            AffineTransform atEl = AffineTransform.getTranslateInstance(r[3],r[4]+r[6]/2);
            atEl.rotate(r[7]);
            
            Rectangle2D.Double sq = new Rectangle2D.Double(-frameBorder,-frameBorder,2*frameBorder,2*frameBorder);
            cur = atEl.createTransformedShape(sq);
            if(cur.contains((double)e.getX(),(double)e.getY()))
                return 1;
            
            sq = new Rectangle2D.Double(r[5]-frameBorder,-frameBorder,2*frameBorder,2*frameBorder);
            cur = atEl.createTransformedShape(sq);
            if(cur.contains((double)e.getX(),(double)e.getY()))
                return 2;
            
            sq = new Rectangle2D.Double(r[5]/2-frameBorder,-r[6]/2-frameBorder,2*frameBorder,2*frameBorder);
            cur = atEl.createTransformedShape(sq);
            if(cur.contains((double)e.getX(),(double)e.getY()))
                return 3;
            
            sq = new Rectangle2D.Double(r[5]/2-frameBorder,r[6]/2-frameBorder,2*frameBorder,2*frameBorder);
            cur = atEl.createTransformedShape(sq);
            if(cur.contains((double)e.getX(),(double)e.getY()))
                return 4;
            
            Ellipse2D.Double el = new Ellipse2D.Double(0,-r[6]/2,r[5],r[6]);
            cur = atEl.createTransformedShape(el);
            if(cur.contains((double)e.getX(),(double)e.getY()))
                return 0;
            
            return -1;
        } else if((int)r[0] == POLY) {
            int x,y;
            int np = 0;
            for(int p = 3; p < r.length; p++) {
                x = (int)r[p];
                p++;
                y = (int)r[p];
                np++;
                
                Rectangle2D.Double sq = new Rectangle2D.Double(x-frameBorder,y-frameBorder,2*frameBorder,2*frameBorder);
                if(sq.contains((double)e.getX(),(double)e.getY()))
                    return np;
            }

            int[] xp = new int[(r.length-3)/2];
            int[] yp = new int[(r.length-3)/2];
            int n = 0;
            for(int p = 3; p < r.length; p++) {
                xp[n] = (int)r[p];
                p++;
                yp[n] = (int)r[p];
                n++;
            }
            Polygon poly = new Polygon(xp,yp,(r.length-3)/2);
            if(poly.contains((double)e.getX(),(double)e.getY()))
                return 0;
            
            return -1;
        } else if((int)r[0] == QUAD) {
            Rectangle2D.Double rect = new Rectangle2D.Double(r[3]-frameBorder,r[4]-frameBorder,2*frameBorder,2*frameBorder);
            if(rect.contains((double)e.getX(),(double)e.getY()))
                return 0;
            
            return -1;
        } else if((int)r[0] == RANGE) {
            double theta = Math.atan((r[6]-r[4])/(r[5]-r[3]));
            if((theta < Math.PI/5 && theta > -Math.PI/5) || mode == HIST || mode == SM_HIST) {
                Rectangle2D.Double sq = new Rectangle2D.Double(r[3]-frameBorder,r[4]-frameBorder,2*frameBorder,2*frameBorder);
                if(sq.contains((double)e.getX(),(double)e.getY()))
                    return 1;
                sq = new Rectangle2D.Double(r[5]-frameBorder,r[4]-frameBorder,2*frameBorder,2*frameBorder);
                if(sq.contains((double)e.getX(),(double)e.getY()))
                    return 2;
                sq = new Rectangle2D.Double(r[3],r[4]-frameBorder,r[5]-r[3],2*frameBorder);
                if(sq.contains((double)e.getX(),(double)e.getY()))
                    return 0;
            } else {
                Rectangle2D.Double sq = new Rectangle2D.Double(r[3]-frameBorder,r[4]-frameBorder,2*frameBorder,2*frameBorder);
                if(sq.contains((double)e.getX(),(double)e.getY()))
                    return 1;
                sq = new Rectangle2D.Double(r[3]-frameBorder,r[6]-frameBorder,2*frameBorder,2*frameBorder);
                if(sq.contains((double)e.getX(),(double)e.getY()))
                    return 2;
                sq = new Rectangle2D.Double(r[3]-frameBorder,r[4],2*frameBorder,r[6]-r[4]);
                if(sq.contains((double)e.getX(),(double)e.getY()))
                    return 0;
            }
            return -1;
        }
        return -1;
    }
    
    public double[] generatePctLabel(Vector regions, int regionEditing) {
		//see also line 1650
        /*double[] reg = (double[])regions.elementAt(regionEditing);
        if((int)reg[0] != QUAD) {
            Vector gate = new Vector();
            int[] di = new int[1]; di[0] = fData[0].index;
            gate.add(new Gate(di, regionEditing,""));
            FACSData[] f0 = new FACSData[1]; f0[0] = fData[0];
            FACSData[] gated = facs.gate(f0,gate);
            double ptg = (double)gated[0].data.length / (double)fData[0].data.length;
            
            double[] pct = { 1, ptg, getCtr((double[])regions.elementAt(regionEditing),0)[0], getCtr((double[])regions.elementAt(regionEditing),0)[1] };
            return pct;
        } else {
            
            double[] pct = { 1, facs.gateQuads(regionEditing,1,fData[0]), getCtr((double[])regions.elementAt(regionEditing),1)[0], getCtr((double[])regions.elementAt(regionEditing),1)[1],
                facs.gateQuads(regionEditing,2,fData[0]), getCtr((double[])regions.elementAt(regionEditing),2)[0], getCtr((double[])regions.elementAt(regionEditing),2)[1],
                facs.gateQuads(regionEditing,3,fData[0]), getCtr((double[])regions.elementAt(regionEditing),3)[0], getCtr((double[])regions.elementAt(regionEditing),3)[1],
                facs.gateQuads(regionEditing,4,fData[0]), getCtr((double[])regions.elementAt(regionEditing),4)[0], getCtr((double[])regions.elementAt(regionEditing),4)[1]};
            return pct;
        }*/
		return null;
    }
    
    public void mouseReleased(MouseEvent e) {
        if(!mouseIn(e) && regionEditing == -1) { //it's an event for the PlotPanel
            /*e.translatePoint((int)getLocation().getX(),(int)getLocation().getY());
            pp.mouseReleased(e);*/
            return;
        }
        if(!draw)
            return;
        if(region != NONE) {
            if(region != POLY) {
                rStartX = -1;
                rStartY = -1;
                if(regionEditing != -1) {
                	if(defaultRegionName == null) {
                		String name = JOptionPane.showInputDialog(null,"Name New Gate",((Region)regions.elementAt(regionEditing)).getName());
                		((Region)regions.elementAt(regionEditing)).setName(name);
                	}
                	
                    Region unT = unTranslate((Region)regions.elementAt(regionEditing));
                    facs.regions.add(unT);
                    //percents.add(generatePctLabel(regions,regionEditing));
                    facs.updatePlots(regionEditing,true);
                }
                regionEditing = -1;
                region = NONE;
                repaint();
                setCursor(null);
            }
        } else if(regionEditing != -1) {
            Region unT = unTranslate((Region)regions.elementAt(regionEditing));
            facs.regions.setElementAt(unT,regionEditing);
            
            Vector gate = new Vector();
            int[] di = new int[1]; di[0] = 0;
            gate.add(new Gate(facs.getFDataNames(di), regionEditing,""));
            /*FACSData[] f0 = new FACSData[1]; f0[0] = fData[0];
            FACSData[] gated = facs.gate(f0,gate);
            double ptg = (double)gated[0].data.length / (double)fData[0].data.length;
            
            double[] pct = { 1, ptg, getCtr((double[])regions.elementAt(regionEditing),0)[0], getCtr((double[])regions.elementAt(regionEditing),0)[1] };
            percents.setElementAt(pct,regionEditing);*/

            facs.updatePlots(regionEditing,true);
        }
    }
    
    public void mouseDragged(MouseEvent e) {
        if(!mouseIn(e) && regionEditing == -1) { //it's an event for the PlotPanel
            /*e.translatePoint((int)getLocation().getX(),(int)getLocation().getY());
            pp.mouseDragged(e);*/
            return;
        } else {
            grabFocus();
        }
        if(!draw)
            return;
        if(region != NONE && rStartX != -1 && rStartY != -1) {
            if(mode != HIST && mode != SM_HIST) {
                if(region == ELLIPSE) {
                    double[] el = ((Region)regions.elementAt(regionEditing)).getVals(fData[0].filename);
                    el[5] = Point2D.distance(rStartX,rStartY,e.getX(),e.getY()); //w
                    el[6] = el[5]/2.0; //h (user can change later
                    el[4] = rStartY - el[6]/2.0; //y (subtract h because refers to bounding rect)
                    el[7] = Math.atan2(e.getY()-rStartY,e.getX()-rStartX);
                    ((Region)regions.elementAt(regionEditing)).setVals(fData[0].filename,el);
                } else if(region == RECT) {
                    double[] r = ((Region)regions.elementAt(regionEditing)).getVals(fData[0].filename);
                    r[5] = Math.max(e.getX() - rStartX,0);
                    r[6] = Math.max(e.getY() - rStartY,0);
                    ((Region)regions.elementAt(regionEditing)).setVals(fData[0].filename,r);
                } else if(region == RANGE) {
                    double[] r = ((Region)regions.elementAt(regionEditing)).getVals(fData[0].filename);
                    r[5] = e.getX();
                    r[6] = e.getY();
                    ((Region)regions.elementAt(regionEditing)).setVals(fData[0].filename,r);
                }
                repaint();
            } else if(region == RANGE) {
                double[] r = ((Region)regions.elementAt(regionEditing)).getVals(fData[0].filename);
                r[5] = e.getX();
                r[6] = e.getY();
                
                double theta = Math.atan((r[6]-r[4])/(r[5]-r[3]));
                boolean horiz = (theta < Math.PI/5 && theta > -Math.PI/5) || mode == HIST || mode == SM_HIST;
                
                if((horiz && r[5] < r[3]) || (!horiz && r[6] < r[4])) { //switch order
                    double t1 = r[5]; double t2 = r[6];
                    r[5] = r[3]; r[6] = r[4];
                    r[3] = t1; r[4] = t2;
                }
                
                ((Region)regions.elementAt(regionEditing)).setVals(fData[0].filename,r);
                repaint();
            }
        } else if(regionEditing != -1) {
            double[] r = ((Region)regions.elementAt(regionEditing)).getVals(fData[0].filename);
            if((int)r[0] == RECT) {
                if(paramEdit == 0) {
                    int deltaX = e.getX()-rStartX;
                    int deltaY = e.getY()-rStartY;
                    r[3] += deltaX;
                    r[4] += deltaY;
                    rStartX = e.getX();
                    rStartY = e.getY();
                    
                    ((Region)regions.elementAt(regionEditing)).setVals(fData[0].filename,r);
                    repaint();
                } else if(paramEdit == 1) { //ul
                    int deltaX = e.getX()-rStartX;
                    int deltaY = e.getY()-rStartY;
                    r[3] += deltaX;
                    r[4] += deltaY;
                    r[5] -= deltaX;
                    r[6] -= deltaY;
                    rStartX = e.getX();
                    rStartY = e.getY();
                    
                    ((Region)regions.elementAt(regionEditing)).setVals(fData[0].filename,r);
                    repaint();
                } else if(paramEdit == 2) { //bl
                    int deltaX = e.getX()-rStartX;
                    int deltaY = e.getY()-rStartY;
                    r[3] += deltaX;
                    r[5] -= deltaX;
                    r[6] += deltaY;
                    rStartX = e.getX();
                    rStartY = e.getY();
                    
                    ((Region)regions.elementAt(regionEditing)).setVals(fData[0].filename,r);
                    repaint();
                } else if(paramEdit == 3) { //ur
                    int deltaX = e.getX()-rStartX;
                    int deltaY = e.getY()-rStartY;
                    r[4] += deltaY;
                    r[5] += deltaX;
                    r[6] -= deltaY;
                    rStartX = e.getX();
                    rStartY = e.getY();
                    
                   ((Region)regions.elementAt(regionEditing)).setVals(fData[0].filename,r);
                    repaint();
                } else if(paramEdit == 4) { //br
                    int deltaX = e.getX()-rStartX;
                    int deltaY = e.getY()-rStartY;
                    r[5] += deltaX;
                    r[6] += deltaY;
                    rStartX = e.getX();
                    rStartY = e.getY();
                    
                    ((Region)regions.elementAt(regionEditing)).setVals(fData[0].filename,r);
                    repaint();
                }
            } else if((int)r[0] == ELLIPSE) {
                if(paramEdit == 0) {
                    int deltaX = e.getX()-rStartX;
                    int deltaY = e.getY()-rStartY;
                    r[3] += deltaX;
                    r[4] += deltaY;
                    rStartX = e.getX();
                    rStartY = e.getY();
                    
                    ((Region)regions.elementAt(regionEditing)).setVals(fData[0].filename,r);
                    repaint();
                } else if(paramEdit == 1) { //left
                    double y2 = r[4]+r[6]/2.0+r[5]*Math.sin(r[7]); //second point of ellipse (not explicitly stored)
                    double x2 = r[3]+r[5]*Math.cos(r[7]);
                    
                    r[3] += e.getX()-rStartX;
                    r[4] += e.getY()-rStartY;
                    
                    r[5] = Point2D.distance(r[3],r[4]+r[6]/2.0,x2,y2);
                    r[7] = Math.atan2(y2-(r[4]+r[6]/2.0),x2-r[3]);
                    
                    rStartX = e.getX();
                    rStartY = e.getY();
                    
                    ((Region)regions.elementAt(regionEditing)).setVals(fData[0].filename,r);
                    repaint();
                } else if(paramEdit == 2) { //right
                    r[5] = Point2D.distance(r[3],r[4]+r[6]/2.0,(double)e.getX(),(double)e.getY()); //w
                    r[7] = Math.atan2((double)e.getY()-(r[4]+r[6]/2.0),(double)e.getX()-r[3]);
                    ((Region)regions.elementAt(regionEditing)).setVals(fData[0].filename,r);
                    repaint();
                } else if(paramEdit == 3) { //top
                    AffineTransform at = AffineTransform.getTranslateInstance(-r[3],-r[4]-r[6]/2);
                    at.rotate(-r[7]);
                    Point2D pOld = at.transform(new Point2D.Double((double)rStartX,(double)rStartY),null);
                    Point2D pNew = at.transform(new Point2D.Double((double)e.getX(),(double)e.getY()),null);
                    double delta = (pNew.getY() - pOld.getY())*2;
                    r[6] -= delta;
                    r[4] += delta/2;
                    
                    rStartX = e.getX();
                    rStartY = e.getY();
                    
                    ((Region)regions.elementAt(regionEditing)).setVals(fData[0].filename,r);
                    repaint();
                } else if(paramEdit == 4) { //bottom
                    AffineTransform at = AffineTransform.getTranslateInstance(-r[3],-r[4]-r[6]/2);
                    at.rotate(-r[7]);
                    Point2D pOld = at.transform(new Point2D.Double((double)rStartX,(double)rStartY),null);
                    Point2D pNew = at.transform(new Point2D.Double((double)e.getX(),(double)e.getY()),null);
                    double delta = (-pNew.getY() + pOld.getY())*2;
                    r[6] -= delta;
                    r[4] += delta/2;
                    
                    rStartX = e.getX();
                    rStartY = e.getY();
                    
                    ((Region)regions.elementAt(regionEditing)).setVals(fData[0].filename,r);
                    repaint();
                }
            } else if((int)r[0] == POLY) {
                if(paramEdit > 0) {
                    r[2*(paramEdit-1)+3] = e.getX();
                    r[2*(paramEdit-1)+4] = e.getY();
                    ((Region)regions.elementAt(regionEditing)).setVals(fData[0].filename,r);
                    repaint();
                } else if(paramEdit == 0) { //move whole shape...
                    int deltaX = e.getX()-rStartX;
                    int deltaY = e.getY()-rStartY;
                    for(int i = 3; i < r.length; i++) {
                        r[i] += deltaX;
                        i++;
                        r[i] += deltaY;
                    }
                    rStartX = e.getX();
                    rStartY = e.getY();
                    
                    ((Region)regions.elementAt(regionEditing)).setVals(fData[0].filename,r);
                    repaint();
                }
            } else if((int)r[0] == RANGE) {
                if(paramEdit == 0) {
                    int deltaX = e.getX()-rStartX;
                    int deltaY = e.getY()-rStartY;
                    r[3] += deltaX;
                    r[4] += deltaY;
                    r[5] += deltaX;
                    r[6] += deltaY;
                    rStartX = e.getX();
                    rStartY = e.getY();
                    
                    ((Region)regions.elementAt(regionEditing)).setVals(fData[0].filename,r);
                    repaint();
                } else if(paramEdit == 1) {
                    double theta = Math.atan((r[6]-r[4])/(r[5]-r[3]));
                    boolean horiz = (theta < Math.PI/5 && theta > -Math.PI/5) || mode == HIST || mode == SM_HIST;
                    if(horiz)
                        r[3] = e.getX();
                    else
                        r[4] = e.getY();
                    
                    theta = Math.atan((r[6]-r[4])/(r[5]-r[3]));
                    horiz = (theta < Math.PI/5 && theta > -Math.PI/5) || mode == HIST || mode == SM_HIST;
                    
                    if((horiz && r[5] < r[3]) || (!horiz && r[6] < r[4])) { //switch order
                        double t1 = r[5]; double t2 = r[6];
                        r[5] = r[3]; r[6] = r[4];
                        r[3] = t1; r[4] = t2;
                        paramEdit = 2;
                    }
                    
                    ((Region)regions.elementAt(regionEditing)).setVals(fData[0].filename,r);
                    repaint();
                } else if(paramEdit == 2) {
                    r[5] = e.getX();
                    r[6] = e.getY();
                    
                    double theta = Math.atan((r[6]-r[4])/(r[5]-r[3]));
                    boolean horiz = (theta < Math.PI/5 && theta > -Math.PI/5) || mode == HIST || mode == SM_HIST;
                    if((horiz && r[5] < r[3]) || (!horiz && r[6] < r[4])) { //switch order
                        double t1 = r[5]; double t2 = r[6];
                        r[5] = r[3]; r[6] = r[4];
                        r[3] = t1; r[4] = t2;
                        paramEdit = 1;
                    }
                    
                    ((Region)regions.elementAt(regionEditing)).setVals(fData[0].filename,r);
                    repaint();
                }
            } else if((int)r[0] == QUAD) {
                if(paramEdit == 0) {
                    r[3] = e.getX();
                    r[4] = e.getY();
                    ((Region)regions.elementAt(regionEditing)).setVals(fData[0].filename,r);
                    repaint();
                }
            }
        }
    }
    
    public void mouseMoved(MouseEvent e) {
        if(!mouseIn(e)) { //it's an event for the PlotPanel
            setCursor(null);
            /*e.translatePoint((int)getLocation().getX(),(int)getLocation().getY());
            pp.mouseMoved(e);*/
            return;
        }
        if(!draw)
            return;
        if(region != NONE) {
            if(e.getX() > axisBorder+frameBorder && e.getX() <= axisBorder+frameBorder+w
                                  && e.getY() > frameBorder+title && e.getY() < frameBorder+title+h) {
                if((mode != HIST && mode != SM_HIST) || region == RANGE) {
                    setCursor(new Cursor(Cursor.CROSSHAIR_CURSOR));
                    if(region == QUAD) {
                        double[] r = { (double)QUAD, param1, param2, (double)e.getX(), (double)e.getY() }; //x, y
                        if(regionEditing == -1) {
                            regionEditing = regions.size();
                            regions.add(new Region(r,makeRName("G"+regionEditing)));
                            rDisplayed.add(new Boolean(true));
							rLocked.add(new Boolean(false));
                        } else {
                        	((Region)regions.elementAt(regionEditing)).setVals(fData[0].filename,r);
                        }
                        repaint();
                    } else if(region == POLY) {
                        if(regionEditing != -1) {
                            double[] r = ((Region)regions.elementAt(regionEditing)).getVals(fData[0].filename);
                            r[r.length-2] = e.getX();
                            r[r.length-1] = e.getY();
							((Region)regions.elementAt(regionEditing)).setVals(fData[0].filename,r);
                            repaint();
                        }
                    }
                }
            } else {
                setCursor(null);
            }
        } else {
            setCursor(null);
        }
    }
    
    public void itemStateChanged(ItemEvent e) {
        if(! (e.getSource() instanceof JCheckBoxMenuItem))
            return;
        JCheckBoxMenuItem cbmi = (JCheckBoxMenuItem)e.getItemSelectable();
        if(cbmi.getActionCommand().equals("smooth")) {
            if(e.getStateChange() == ItemEvent.SELECTED)
                setParams(param1,param2,SM_HIST);
            else
                setParams(param1,param2,HIST);
            old1D = mode;
        } else if(cbmi.getActionCommand().equals("second")) {
            if(e.getStateChange() == ItemEvent.SELECTED)
                hSecondMax = true;
            else
                hSecondMax = false;
            setParams(param1,param2,mode);
        } else if(cbmi.getActionCommand().equals("outliers")) {
            if(e.getStateChange() == ItemEvent.SELECTED)
                cShowOutliers = true;
            else
                cShowOutliers = false;
            setParams(param1,param2,mode);
        } else if(cbmi.getActionCommand().equals("dsmooth")) {
            if(e.getStateChange() == ItemEvent.SELECTED)
                smooth = true;
            else
                smooth = false;
            setParams(param1,param2,mode);
        }
    }
    
    public String getToolTipText(MouseEvent e) {
        if(!draw)
            return "";
        int region = regionHere(e);
        if(region != -1) {
            return ((Region)facs.regions.elementAt(region)).getName();
        } else {
            String data = "";
            for(int i = 0; i < fData.length; i++) {
                data += fData[i].name;
                if(i != fData.length-1)
                    data += ", ";
            }
            return data;
        }
    }
}

class Gate {
    //int[] dataIndices;
	String[] data;
    int region;
    String name;
    
    public Gate(String[] d, int r, String name) {
        data= d;
        region = r;
        this.name = name;
    }
    
    public String toString() {
        return name;
    }
    
    public String getWritable() {
        String w = region+"|";
        for(int i = 0; i < data.length; i++) {
            w += data[i]+"|";
        }
        return w.substring(0,w.length()-1);
    }
    
    public static Gate fromWritable(String s, String name) {
        String[] pcs = s.split("\\|");
        int region = new Integer(pcs[0]).intValue();
        String[] di = new String[pcs.length-1];
        for(int i = 1; i < pcs.length; i++) {
            //di[i-1] = new Integer(pcs[i]).intValue();
			di[i-1] = pcs[i];
        }
        return new Gate(di, region, name);
    }
}

class RegionsDialog extends JDialog implements ActionListener {
    RVector regions;
    Vector rDisplayed, rLocked;
    FParameter[] params;
    boolean okay;
    JTable table;
    boolean showVis;
	String file;
    
    public RegionsDialog(boolean showVis, String file, RVector regions, Vector rDisplayed, Vector rLocked, FParameter[] params) {
        super((JFrame)null, "Show/Hide Gates", true);
        this.regions = regions;
        this.rDisplayed = rDisplayed;
		this.rLocked = rLocked;
        this.params = params;
        this.showVis = showVis;
		this.file = file;
        okay = false;
        
        getContentPane().setLayout(new BoxLayout(getContentPane(), BoxLayout.Y_AXIS));
        
        RegionsTableModel rtm = new RegionsTableModel(showVis,file,regions,rDisplayed,rLocked,params);
        table = new JTable(rtm);
        table.setPreferredScrollableViewportSize(new Dimension(650,150));
        
        JScrollPane sp = new JScrollPane(table);
        getContentPane().add(sp);
        
        JPanel pBut = new JPanel();
        JButton bOk = new JButton("OK");
        bOk.addActionListener(this);
        bOk.setActionCommand("ok");
        pBut.add(bOk);
        JButton bCancel = new JButton("Cancel");
        bCancel.addActionListener(this);
        bCancel.setActionCommand("cancel");
        pBut.add(bCancel);
        getContentPane().add(pBut);
        
        pack();
        setVisible(true);
    }
    
    public void actionPerformed(ActionEvent e) {
        if(e.getActionCommand() == "ok") {
            okay = true;
            setVisible(false);
        }
        if(e.getActionCommand() == "cancel") {
            setVisible(false);
        }
    }
}

class RegionsTableModel extends AbstractTableModel {
    RVector regions;
    Vector rDisplayed, rLocked;
    FParameter[] params;
    Vector cols;
    boolean showVis;
	String file;
    
    public RegionsTableModel(boolean showVis, String file, RVector regions, Vector rDisplayed, Vector rLocked, FParameter[] params) {
        this.regions = regions;
        this.rDisplayed = rDisplayed;
		this.rLocked = rLocked;
        this.params = params;
        this.showVis = showVis;
		this.file = file;
        
        cols = new Vector();
        if(showVis)
            cols.addElement("Visible");
		cols.addElement("Locked");
		cols.addElement("Global");
        cols.addElement("Name");
        cols.addElement("Type");
        cols.addElement("Parameter 1");
        cols.addElement("Parameter 2");
    }
    
    public int getColumnCount() {
        return cols.size();
    }
    
    public int getRowCount() {
        return regions.size();
    }
    
    public String getColumnName(int col) {
        return (String)cols.elementAt(col);
    }
    
    public Object getValueAt(int row, int col) {
        if(!showVis)
            col++;
        if(col == 0)
            return (Boolean)rDisplayed.elementAt(row);
		if(col == 1)
            return (Boolean)rLocked.elementAt(row);
		if(col == 2)
            return new Boolean(((Region)regions.elementAt(row)).isGlobal(file));
        if(col == 3)
            return ((Region)regions.elementAt(row)).getName();
        if(col == 4)
            return getType((int)((Region)regions.elementAt(row)).getVals(file)[0]);
        if(col == 5)
            return getParam(((Region)regions.elementAt(row)).getVals(file),1);
        if(col == 6) {
            return getParam(((Region)regions.elementAt(row)).getVals(file),2);
        }
		
        return "";
    }
    
    public String getParam(double[] gate, int num) {
        if((int)gate[0] != Plot.RANGE)
            return params[(int)gate[num]].lname;
        
        
        double theta = Math.atan((gate[6]-gate[4])/(gate[5]-gate[3]));
        if((theta < Math.PI/5 && theta > -Math.PI/5) || gate[2] == -1) {
            if(num == 1)
                return params[(int)gate[1]].lname;
            else
                return "";
        } else {
            if(num == 1)
                return params[(int)gate[2]].lname;
            else
                return "";
        }
    }
    
    public String getType(int i) {
        i = Math.abs(i);
        if(i == Plot.RECT)
            return "Rectangle";
        if(i == Plot.ELLIPSE)
            return "Ellipse";
        if(i == Plot.QUAD)
            return "Quadrants";
        if(i == Plot.POLY)
            return "Polygon";
        if(i == Plot.RANGE)
            return "Range";
        return "error";
    }
    
    public Class getColumnClass(int c) {
        return getValueAt(0,c).getClass();
    }
    
    public boolean isCellEditable(int row, int col) {
        if(!showVis)
            col++;
        if(col < 4)
            return true;
        return false;
    }
    
    public void setValueAt(Object value, int row, int col) {
        if(!showVis)
            col++;
        if(col == 0) {
            rDisplayed.setElementAt(value,row);
        }
        if(col == 3) {
            ((Region)regions.elementAt(row)).setName((String)value);
        }
		if(col == 2) {
			((Region)regions.elementAt(row)).setGlobal(file,((Boolean)value).booleanValue());
			rLocked.setElementAt(new Boolean(false),row);
			fireTableCellUpdated(row,1);
		}
		if(col == 1) {
			rLocked.setElementAt(value,row);
		}
        fireTableCellUpdated(row,col);
    }
}

class PlotKeyDialog extends JDialog {
	public PlotKeyDialog(PlotKey pk) {
		super((JFrame)null,"Plot Key",false);

		getContentPane().add(pk);
			
		pack();
		setVisible(true);
	}
}

class PlotKey extends JComponent implements MouseListener, MouseMotionListener {
    //PlotPanel pp;
    Hashtable datas;
    Vector dnames;
    Plot plot;
    Font f;
    FontMetrics fm;
    boolean selected;
    
    public PlotKey(Vector dnames, Hashtable datas, Plot plot) {
        //this.pp = pp;
        this.dnames = dnames;
        this.datas = datas;
        this.plot = plot;
        f = new Font("Serif",Font.PLAIN, 14);
        fm = null;
        selected = false;
        addMouseListener(this);
        addMouseMotionListener(this);

		new PlotKeyDialog(this);
    }
    
    public void mouseClicked(MouseEvent me) {
        chooseColor(me);
    }
    
    public void mousePressed(MouseEvent me) {
        /*me.translatePoint((int)getLocation().getX(),(int)getLocation().getY());
        pp.mousePressed(me);*/
    }
    
    public void mouseDragged(MouseEvent me) {
        /*me.translatePoint((int)getLocation().getX(),(int)getLocation().getY());
        pp.mouseDragged(me);*/
    }

    public void chooseColor(MouseEvent me) {
        if(me.getX() >= 3 && me.getX() <= 3+fm.getHeight()) {
            for(int i = 0; i < dnames.size(); i++) {
                String s = (String)dnames.elementAt(i);
                if(me.getY() >= (2+i*(fm.getHeight()+2)) && me.getY() <= (2+(i+1)*(fm.getHeight()+2))) {
                    Color c = JColorChooser.showDialog(this, "Choose Display Color", (Color)datas.get(s));
                    if(c != null) {
                        datas.put(s,c);
                        repaint();
                        plot.colors[i] = c;
                        plot.setParams(plot.param1, plot.param2, plot.mode);
                    }
                    return;
                }
            }
        }
        /*me.translatePoint((int)getLocation().getX(),(int)getLocation().getY());
        pp.mouseClicked(me);*/
    }
    
    public void paintComponent(Graphics g) {
        Graphics2D g2 = (Graphics2D)g;
        
        g2.setFont(f);
        fm = g2.getFontMetrics();
        
        setSize(getPreferredSize());
        
        g2.setBackground(Color.white);
        g2.setColor(Color.black);
        g2.drawRect(0,0,getWidth()-1,getHeight()-1);
        if(selected)
            g2.drawRect(1,1,getWidth()-3,getHeight()-3);
        
        double y = 2.0;
        for(int i = 0; i < dnames.size(); i++) {
            String s = (String)dnames.elementAt(i);
            Color c = (Color)datas.get(s);
            g2.setColor(c);
            g2.fill(new Rectangle2D.Double(3, y+1, fm.getHeight(), fm.getHeight()));
            g2.setColor(Color.black);
            g2.drawString(s,fm.getHeight()+3,(int)y+fm.getAscent());
            y += fm.getHeight()+2;
        }
    }
    
    public Dimension getPreferredSize() {
        if(fm == null)
            return new Dimension(10,10);
        
        Enumeration e = datas.keys();
        int max = 0;
        while(e.hasMoreElements()) {
            String s = (String)e.nextElement();
            if(fm.stringWidth(s) > max)
                max = fm.stringWidth(s);
        }
        return new Dimension(max+fm.getHeight()+10,4+(fm.getHeight()+2)*datas.size());
    }
    
    public Dimension getMinimumSize() {
        return getPreferredSize();
    }
    
    public Dimension getMaximumSize() {
        return getPreferredSize();
    }
    
    public void mouseEntered(MouseEvent e) {
    }
    
    public void mouseExited(MouseEvent e) {
    }
    
    public void mouseReleased(MouseEvent me) {
        /*me.translatePoint((int)getLocation().getX(),(int)getLocation().getY());
        pp.mouseReleased(me);*/
    }
    
    public void mouseMoved(MouseEvent e) {
    } 
}

class DataDialog extends JDialog implements ActionListener, ListSelectionListener {
    boolean okay;
    FACSData[] sel;
    Vector data;
    
    Vector added;
    Vector gates;
    
    String[] names;
    
    JList list, listSel, listGates;
    DefaultListModel dlmSel;
    
    int oldsel;
    Hashtable selGates;
    
    public DataDialog(JComponent comp, FACSData[] sel, Vector data, Vector gates) {
        super((JFrame)null,"Choose Data...",true);
        setLocationRelativeTo(comp);
        
        this.sel = sel;
        this.data = data;
        okay = false;
        added = new Vector();
        this.gates = gates;
        selGates = new Hashtable();
        oldsel = -1;
        
        for(int d = 0; d < data.size(); d++) {
            Vector v = new Vector();
            for(int i = 0; i < gates.size(); i++) {
                Gate g = (Gate)gates.elementAt(i);
                FACSData fsel = (FACSData)data.elementAt(d);
                if(arrayContains(g.data, fsel.filename)) {
                    v.add(new Integer(i));
                }
            }
            selGates.put(new Integer(d),v);
        }
        
        names = new String[data.size()];
        FACSData f;
        for(int i = 0; i < data.size(); i++) {
            f = (FACSData)data.elementAt(i);
            names[i] = f.getName();
        }
        
        getContentPane().setLayout(new BoxLayout(getContentPane(), BoxLayout.Y_AXIS));
        
        
        JPanel pLists = new JPanel();
        
        list = new JList(names);
        list.setSelectionMode(ListSelectionModel.MULTIPLE_INTERVAL_SELECTION);
        list.setVisibleRowCount(-1);
        JScrollPane spList = new JScrollPane(list);
        spList.setPreferredSize(new Dimension(120,170));
        spList.setBorder(BorderFactory.createTitledBorder("All Loaded"));
        pLists.add(spList);
        
        JPanel pAdd = new JPanel();
        pAdd.setLayout(new GridLayout(2,1));
        JButton bAdd = new JButton("-->");
        bAdd.setActionCommand("add");
        bAdd.addActionListener(this);
        pAdd.add(bAdd);
        JButton bRem = new JButton("<--");
        bRem.setActionCommand("rem");
        bRem.addActionListener(this);
        pAdd.add(bRem);
        
        pLists.add(pAdd);
        
        dlmSel = new DefaultListModel();
        listSel = new JList(dlmSel);
        if(sel != null) {
            //int[] selI = new int[sel.length];
            for(int i = 0; i < sel.length; i++) {
                added.add(new Integer(sel[i].index));
                dlmSel.addElement(names[sel[i].index]);
            }
        }
        listSel.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
        listSel.setVisibleRowCount(-1);
        listSel.addListSelectionListener(this);
        JScrollPane spListSel = new JScrollPane(listSel);
        spListSel.setPreferredSize(new Dimension(120,170));
        spListSel.setBorder(BorderFactory.createTitledBorder("This Plot"));
        pLists.add(spListSel);
        
        listGates = new JList(gates);
        listGates.setSelectionMode(ListSelectionModel.MULTIPLE_INTERVAL_SELECTION);
        listGates.setVisibleRowCount(-1);
        JScrollPane spListGates = new JScrollPane(listGates);
        spListGates.setPreferredSize(new Dimension(120,170));
        spListGates.setBorder(BorderFactory.createTitledBorder("Gates"));
        pLists.add(spListGates);
        
        getContentPane().add(pLists);
        
        JPanel pBut = new JPanel();
        JButton bOk = new JButton("OK");
        bOk.setActionCommand("ok");
        bOk.addActionListener(this);
        pBut.add(bOk);
        JButton bC = new JButton("Cancel");
        bC.setActionCommand("cancel");
        bC.addActionListener(this);
        pBut.add(bC);
        getContentPane().add(pBut);
        
        MouseListener mouseListener1 = new MouseAdapter() { //double-click
            public void mouseClicked(MouseEvent e) {
                if (e.getClickCount() == 2) {
                    add();
                }
            }
        };
        list.addMouseListener(mouseListener1);
        
        listSel.setSelectedIndex(0);
        
        pack();
        setVisible(true);
    }

	public static boolean arrayContains(String[] array, String item) {
        for(int i = 0; i < array.length; i++) {
            if(array[i].equals(item))
                return true;
        }
        return false;
    }
    
    public void add() {
        for(int i = 0; i < list.getSelectedIndices().length; i++) {
            if(!added.contains(new Integer(list.getSelectedIndices()[i]))) {
                added.add(new Integer(list.getSelectedIndices()[i]));
                dlmSel.addElement(list.getSelectedValues()[i]);
            }
        }
    }
    
    public void rem() {
        added.removeElementAt(listSel.getSelectedIndex());
        dlmSel.removeElement(listSel.getSelectedValue());
    }
    
    public void actionPerformed(ActionEvent e) {
        if(e.getActionCommand().equals("ok")) {
            oldsel = listSel.getSelectedIndex();
            if(oldsel != -1)
                selGates.put((Integer)added.elementAt(oldsel),makeVector(listGates.getSelectedIndices()));
            
            Vector gates2 = new Vector(); //only include the gates that have data for them
            for(int i = 0; i < gates.size(); i++) {
                Gate g = (Gate)gates.elementAt(i);
                Vector v = new Vector();
                for(int j = 0; j < data.size(); j++) { //so the idea is, we scan all data sets to collect a list of gates, then organize by gate
                    Vector d = (Vector)selGates.get(new Integer(j));
                    if(d.contains(new Integer(i)))
                        v.add(((FACSData)data.elementAt(j)).filename);
                }
                g.data = makeSArray(v);
                gates2.add(g);
            }
            
            gates = gates2;
            
            okay = true;
            setVisible(false);
        } else if(e.getActionCommand().equals("cancel")) {
            setVisible(false);
        } else if(e.getActionCommand().equals("add")) {
            add();
        } else if(e.getActionCommand().equals("rem")) {
            rem();
        }
    }
    
    public void valueChanged(ListSelectionEvent lse) {
        if(lse.getValueIsAdjusting())
            return;
        
        //save old vals
        if(oldsel != -1 && oldsel < added.size()) {
            /*for(int i = 0; i < listGates.getSelectedIndices().length; i++) {
                Gate g = (Gate)gates.elementAt(i);
                FACSData fsel = (FACSData)data.elementAt(((Integer)added.elementAt(oldsel)).intValue());
                if(!FACSManip.arrayContains(g.dataIndices, fsel.index)) {
                    int[] di = new int[g.dataIndices.length+1];
                    for(int d = 0; d < g.dataIndices.length; d++) {
                        di[d] = g.dataIndices[d];
                    }
                    di[g.dataIndices.length] = fsel.index;
                }
            }*/
            
            selGates.put((Integer)added.elementAt(oldsel),makeVector(listGates.getSelectedIndices()));
        }
        
        oldsel = listSel.getSelectedIndex();
        
        //load new vals
        Vector v = new Vector();
        if(oldsel != -1)
            v = (Vector)selGates.get((Integer)added.elementAt(oldsel));
        listGates.setSelectedIndices(makeArray(v));
    }
    
    public Vector makeVector(int[] indices) {
        Vector v = new Vector();
        for(int i = 0; i < indices.length; i++) {
            v.add(new Integer(indices[i]));
        }
        return v;
    }

	public int[][] makeArray(Vector in,int nParams) {
        int[][] out = new int[in.size()][nParams];
        for(int i = 0; i < in.size(); i++) {
            out[i] = (int[])in.elementAt(i);
        }
        return out;
    }
    
    public static int[] makeArray(Vector in) {
        int[] out = new int[in.size()];
        for(int i = 0; i < in.size(); i++) {
            out[i] = ((Integer)in.elementAt(i)).intValue();
        }
        return out;
    }

	public static String[] makeSArray(Vector in) {
        String[] out = new String[in.size()];
        for(int i = 0; i < in.size(); i++) {
            out[i] = (String)in.elementAt(i);
        }
        return out;
    }
}